"""
msgbase package for x/84, https://github.com/jquast/x84
"""
import logging
import datetime
from x84.bbs.dbproxy import DBProxy

MSGDB = 'msgbase'
TAGDB = 'tags'

# pylint: disable=C0103
#        Invalid name "logger" for type constant (should match
logger = logging.getLogger()


def get_msg(idx=0):
    """
    Return Msg record instance by index ``idx``.
    """
    return DBProxy(MSGDB)['%d' % (idx,)]


def list_msgs(tags=None):
    """
    Return set of Msg keys matching 1 or more ``tags``, or all.
    """
    if tags is not None and 0 != len(tags):
        msgs = set()
        db_tag = DBProxy(TAGDB)
        for tag in (_tag for _tag in tags if _tag in db_tag):
            msgs.update(db_tag[tag])
        return msgs
    return set([int(key) for key in DBProxy(MSGDB).keys()])


def list_tags():
    """
    Return set of available tags.
    """
    return [_tag.decode('utf8') for _tag in DBProxy(TAGDB).keys()]


class Msg(object):
    """
    the Msg object is record spec for messages held in the msgbase.
    It contains many default properties to describe a conversation:

    'creationtime', the time the message was initialized

    'author', 'recipient', 'subject', and 'body' are envelope parameters.

    'read' becomes a list of handles that have viewed a public message, or a
    single time the message was read by the addressed for private messages.

    'tags' is for use with message groupings, containing a list of strings that
    other messages may share in relation.

    'parent' points to the message this message directly refers to, and
    'threads' points to messages that refer to this message. 'parent' must be
    explicitly set, but children are automaticly populated into 'threads' of
    messages replied to through the send() method.
    """
    # pylint: disable=R0902
    #         Too many instance attributes
    idx = None

    @property
    def ctime(self):
        """
        M.ctime() --> datetime

        Datetime message was instantiated
        """
        return self._ctime

    @property
    def stime(self):
        """
        M.stime() --> datetime

        Datetime message was saved to database
        """
        return self._stime

    def __init__(self, recipient=None, subject=u'', body=u''):
        from x84.bbs.session import getsession
        self._ctime = datetime.datetime.now()
        self._stime = None
        self.author = getsession().handle
        self.recipient = recipient
        self.subject = subject
        self.body = body
        self.tags = set()
        # reply-to tracking
        self.children = set()
        self.parent = None

    def save(self):
        """
        Save message in 'Msgs' sqlite db, and record index in 'tags' db.
        """
        db_msg = DBProxy(MSGDB)
        new = self.idx is None or self._stime is None
        # persist message record to MSGDB
        db_msg.acquire()
        if new:
            self.idx = max([int(key) for key in db_msg.keys()] or [-1]) + 1
            self._stime = datetime.datetime.now()
            new = True
        db_msg['%d' % (self.idx,)] = self
        db_msg.release()

        # persist message idx to TAGDB
        db_tag = DBProxy(TAGDB)
        db_tag.acquire()
        for tag, msgs in db_tag.iteritems():
            if tag in self.tags and not self.idx in msgs:
                msgs.add(self.idx)
                db_tag[tag] = msgs
                logger.info(u"msg %s tagged '%s'", self.idx, tag,)
            elif tag not in self.tags and self.idx in msgs:
                msgs.remove(self.idx)
                db_tag[tag] = msgs
                logger.info(u"msg %s untagged '%s'", self.idx, tag,)
        for tag in [_tag for _tag in self.tags if not _tag in db_tag]:
            db_tag[tag] = set([self.idx])
        db_tag.release()

        # persist message as child to parent;
        if not hasattr(self, 'parent'):
            self.parent = None
        assert self.parent not in self.children
        if self.parent is not None:
            parent_msg = get_msg(self.parent)
            if not hasattr(parent_msg, 'children'):
                parent_msg.children = set(
                )  # intermediary conversion; deleteme
            parent_msg.children.add(self.idx)
            parent_msg.save()

        logger.info(u"saved %s%s%s, addressed to '%s'.",
                    'new ' if new else u'',
                    'public ' if 'public' in self.tags else u'',
                    'message' if self.parent is None else u'reply',
                    self.recipient,)



"""
Terminal handler for x/84 bbs.  http://github.com/jquast/x84
"""
import threading
import logging
import socket
import time
import re

from x84.blessings import Terminal as BlessingsTerminal

TERMINALS = dict()

class Terminal(BlessingsTerminal):
    def __init__(self, kind, stream, rows, columns):
        self.rows = rows
        self.columns = columns
        BlessingsTerminal.__init__(self, kind, stream)

    def kbhit(self, timeout=0):
        from x84.bbs import getsession
        val = getsession().read_event('input', timeout)
        if val is not None:
            getsession()._buffer['input'].append(val)
            return True
        return False

    def getch(self):
        from x84.bbs import getsession
        return getsession().read_event('input')

    def _height_and_width(self):
        return (self.rows, self.columns)

    @property
    def is_a_tty(self):
        return True

def init_term(out_queue, lock, env):
    """
    curses is initialized using the value of 'TERM' of dictionary env,
    as well as a starting window size of 'LINES' and 'COLUMNS'.

    A blessings-abstracted curses terminal is returned.
    """
    from x84.bbs import ini
    from x84.bbs.ipc import IPCStream
    log = logging.getLogger()
    ttype = env.get('TERM', 'unknown')
    if (ttype == 'ansi'
            and ini.CFG.get('system', 'termcap-ansi') != 'no'):
        # special workaround for systems with 'ansi-bbs' termcap,
        # translate 'ansi' -> 'ansi-bbs'
        # http://wiki.synchro.net/install:nix?s[]=termcap#terminal_capabilities
        env['TERM'] = ini.CFG.get('system', 'termcap-ansi')
        log.debug('TERM %s transliterated to %s' % ( ttype, env['TERM'],))
    elif (ttype == 'unknown'
            and ini.CFG.get('system', 'termcap-unknown') != 'no'):
        # instead of using 'unknown' as a termcap definition, try to use
        # the most broadly capable termcap possible, 'vt100', configurable with
        # default.ini
        env['TERM'] = ini.CFG.get('system', 'termcap-unknown')
        log.debug('TERM %s transliterated to %s' % ( ttype, env['TERM'],))
    else:
        log.debug('TERM is %s' % (ttype,))
    stream = IPCStream(out_queue, lock)
    return Terminal(env.get('TERM', 'unknown'), stream,
            int(env.get('LINES', '24')),
            int(env.get('COLUMNS', '80'),))


def mkipc_rlog(out_queue):
    """
    Remove any existing handlers of the current process, and
    re-address the root logging handler to an IPC output event queue
    """
    from x84.bbs.ipc import IPCLogHandler
    root = logging.getLogger()
    for _hdlr in root.handlers:
        root.removeHandler(_hdlr)
    hdlr = IPCLogHandler(out_queue)
    root.addHandler(hdlr)
    return hdlr


class TerminalProcess():
    """
    Class record for tracking global processes and their
    various attributes. These are stored using register() and unregister(),
    and retrieved using terminals().
    """
    # pylint: disable=R0903
    #         Too few public methods
    def __init__(self, client, iqueue, oqueue, lock):
        from x84.bbs.ini import CFG
        self.client = client
        self.iqueue = iqueue
        self.oqueue = oqueue
        self.lock = lock
        self.timeout = CFG.getint('system', 'timeout')

    @property
    def sid(self):
        """ Returns session id. """
        return self.client.addrport()


def register(tty):
    """
    Register a global instance of TerminalProcess
    """
    TERMINALS[tty.sid] = tty


def flush_queue(queue):
    """
    Seeks any remaining events in queue, used before closing
    to prevent zombie processes with IPC waiting to be picked up.
    """
    logger = logging.getLogger()
    try:
        while queue.poll():
            event, data = queue.recv()
            if event == 'logger':
                logger.handle(data)
    except (EOFError, IOError) as err:
        logger.debug(err)


def unregister(tty):
    """
    Unregister a Terminal, described by its telnet.TelnetClient,
    input and output Queues, and Lock.
    """
    logger = logging.getLogger()
    try:
        flush_queue(tty.oqueue)
        tty.oqueue.close()
        tty.iqueue.close()
    except (EOFError, IOError) as err:
        logger.exception(err)
    if tty.client.active:
        # signal tcp socket to close
        tty.client.deactivate()
    del TERMINALS[tty.sid]
    logger.debug('%s: unregistered', tty.client.addrport())


def terminals():
    """
    Returns a list of tuple (telnet.TelnetClient,
        input Queue, output Queue, Lock).
    """
    return TERMINALS.items()


# pylint: disable=R0913
#         Too many arguments (6/5)
def start_process(inp_queue, out_queue, sid, env, lock, binary=False):
    """
    A multiprocessing.Process target. Arguments:
        inp_queue and out_queue: multiprocessing.Queue
        sid: string describing session source (fe. IP address & Port)
        env: dictionary of client environment variables (requires 'TERM')
        binary: If client accepts BINARY, assume utf8 session encoding.
    """
    import x84.bbs.ini
    import x84.bbs.session
    # terminals of these types are forced to 'cp437' encoding,
    # we could also more safely assume iso8859-1, which is the
    # correct default encoding of those terminals, but we explicitly
    # send the 'set graphics character code' attribute for cp437.
    #
    # If they don't honor it, they don't get to see the art correctly.
    cp437_ttypes = ('unknown', 'ansi', 'ansi-bbs', 'vt100',)

    # root handler has dangerously forked file descriptors.
    # replace with ipc 'logger' events so that only the main
    # process is responsible for logging.
    hdlr = mkipc_rlog(out_queue)
    # initialize blessings terminal based on env's TERM.
    term = init_term(out_queue, lock, env)
    # negotiate encoding; terminals with BINARY mode are utf-8
    encoding = x84.bbs.ini.CFG.get('session', 'default_encoding')
    if env.get('TERM', 'unknown') in cp437_ttypes:
        encoding = 'cp437'
    elif env.get('TERM', 'unknown').startswith('vt'):
        encoding = 'cp437'
    elif binary:
        encoding = 'utf8'
    # spawn and begin a new session
    session = x84.bbs.session.Session(
        term, inp_queue, out_queue, sid, env, lock, encoding)
    # copy session ptr to logger handler for 'handle' emit logging
    hdlr.session = session
    # run session
    session.run()
    # signal engine to shutdown subprocess
    out_queue.send(('exit', None))


def on_naws(client):
    """
    On a NAWS event, check if client is yet registered in registry and send
    the input event queue a 'refresh' event. This is the same thing as ^L
    to the 'userland', but should indicate also that the window sizes are
    checked`.
    """
    for _sid, tty in terminals():
        if client == tty.client:
            columns = int(client.env['COLUMNS'])
            rows = int(client.env['LINES'])
            tty.iqueue.send(('refresh', ('resize', (columns, rows),)))
            return True


class ConnectTelnet (threading.Thread):
    """
    Accept new Telnet Connection and negotiate options.
    """
    # this all gets much better using "telnetlib3" tulip/Futures, this is
    # a pretty poor implementation, but necessary for correct "TERM" before
    # going through the trouble of spawning a new process -- that process
    # may only call curses.setupterm() once per process, so it is imperative
    # to make a "good call"; which is a combination of many several factors
    TIME_NEGOTIATE = 2.50  # wait 2500ms on-connect
    TIME_WAIT_STAGE = 3.50  # wait upto 3500ms foreach stage of negotiation :(
    TIME_WAIT_SILENT = 2.50  # wait 2500ms after silence
    TIME_POLL = 0.15
    TTYPE_UNDETECTED = 'unknown'
    WINSIZE_TRICK = (
        ('vt100', ('\x1b[6n'), re.compile(chr(27) + r"\[(\d+);(\d+)R")),
        ('sun', ('\x1b[18t'), re.compile(chr(27) + r"\[8;(\d+);(\d+)t"))
    )  # see: xresize.c from X11.org
    DA_REPLIES = (
        ('\x1b[?1;0c', 'vt100', ''),
        ('\x1b[?1;1c', 'vt100', 'STP'),
        ('\x1b[?1;3c', 'vt100', 'STP;AVO'),
        ('\x1b[?1;4c', 'vt100', 'GPO'),
        ('\x1b[?1;5c', 'vt100', 'STP;GPO'),
        ('\x1b[?1;6c', 'vt100', 'AVO;GPO'),
        ('\x1b[?1;7c', 'vt100', 'STP;AVO;GPO'),
        ('\x1b[?1;11c', 'vt100', 'PP;AVO'),
        ('\x1b[?1;15c', 'vt100', 'PP;GPO;AVO'),
        ('\x1b[?4;2c', 'vt132', 'AVO'),
        ('\x1b[?4;3c', 'vt132', 'AVO;STP'),
        ('\x1b[?4;6c', 'vt132', 'GPO;AVO'),
        ('\x1b[?4;7c', 'vt132', 'GPO;AVO;STP'),
        ('\x1b[?4;11c', 'vt132', 'PP;AVO'),
        ('\x1b[?4;15c', 'vt132', 'PP;GPO;AVO'),
        ('\x1b[?6c', 'vt102', ''),
        ('\x1b[?1;2c', 'vt102', 'AVO'),
        ('\x1b[?7c', 'vt131', ''),
        ('\x1b[?12;5c', 'vt125', ''),
        ('\x1b[?12;7c', 'vt125', 'AVO'),
        ('\x1b[?62;1;2;4;6;8;9;15c', 'vt220', ''),
        ('\x1b[?63;1;2;8;9c', 'vt320', ''),
        ('\x1b[?63;1;2;4;6;8;9;15c', 'vt320', ''),
    ) # see: report.c from vttest



    def __init__(self, client):
        """
        client is a telnet.TelnetClient instance.
        """
        self.client = client
        threading.Thread.__init__(self)

    def _spawn_session(self):
        """
        Spawn a subprocess, avoiding GIL and forcing all shared data over a
        Queue. Previous versions of x/84 and prsv were single process,
        thread-based, and shared variables.

        All IPC communication occurs through the bi-directional queues.  The
        server end (engine.py) polls the out_queue, and places results
        and input events into the inp_queue, while the client end (session.py),
        polls the inp_queue, and places output into out_queue.
        """
        logger = logging.getLogger()
        if not self.client.active:
            logger.debug('session aborted; socket was closed.')
            return
        from multiprocessing import Process, Pipe, Lock
        from x84.telnet import BINARY
        inp_recv, inp_send = Pipe(duplex=False)
        out_recv, out_send = Pipe(duplex=False)
        lock = Lock()
        is_binary = (self.client.check_local_option(BINARY)
                     and self.client.check_remote_option(BINARY))
        child_args = (inp_recv, out_send, self.client.addrport(),
                      self.client.env, lock, is_binary)
        logger.debug('starting session')
        proc = Process(target=start_process, args=child_args)
        proc.start()
        tty = TerminalProcess(self.client, inp_send, out_recv, lock)
        register(tty)

    def banner(self):
        """
        This method is called after the connection is initiated.

        This routine happens to communicate with a wide variety of network
        scanners when listening on the default port on a public IP address.
        """
        logger = logging.getLogger()
        mrk_bytes = self.client.bytes_received
        # According to Roger Espel Llima (espel@drakkar.ens.fr), you can
        #   have your server send a sequence of control characters:
        # (0xff 0xfb 0x01) (0xff 0xfb 0x03) (0xff 0xfd 0x0f3).
        #   Which translates to:
        # (IAC WILL ECHO) (IAC WILL SUPPRESS-GO-AHEAD)
        # (IAC DO SUPPRESS-GO-AHEAD).
        self.client.request_will_echo()
        self.client.request_will_sga()
        self.client.request_do_sga()
        # add DO & WILL BINARY, for utf8 input/output.
        self.client.request_do_binary()
        self.client.request_will_binary()
        self.client.request_do_ttype()
        # and do naws, please.
        self.client.request_do_naws()
        self.client.socket_send()  # push

        if not self.client.active:
            return
        # wait at least 1.25s for the client to speak. If it doesn't,
        # we try only TTYPE. If it fails to report that, we forget
        # the rest.
        logger.debug('pausing for negotiation')
        st_time = time.time()
        while ((0 == self.client.bytes_received
            or mrk_bytes == self.client.bytes_received)
               and time.time() - st_time < self.TIME_NEGOTIATE
               and self.client.active):
            self.client.send_str(chr(0))  # send NUL; keep scanners with us,
            self.client.socket_send()  # push
            time.sleep(self.TIME_POLL)
        mrk_bytes = self.client.bytes_received
        self._try_ttype()
        #if ((0 == self.client.bytes_received
        #          or mrk_bytes == self.client.bytes_received)
        #        or self.client.env.get('TERM', 'unknown')):
        #    # having not received a single byte, we opt out of the
        #    # negotiation program. Usually, the connecting client is
        #    # a scanner; the equivalent of:
        #    #  (printf "root\n"; sleep 5) | nc host > log.txt
        #    logger.info('Dumb terminal detected; no further negotiation.')
        #    return

        self._try_env()
        # this will set Term.kind if -still- undetected,
        # or otherwise overwrite it if it is detected different.
        # First, telnet TTYPE is explicitly requested. If no
        # reply is found, then a dec terminal attributes
        # request is made, and a vt* terminal type is set.

        # XXX Disabled, was not found to be very useful.
        # If this fails, or response is amiguous ('vt100'), try for
        # extended dec types. If not, just get them anyway.
        #self._try_device_attributes()

        # check for naws reply. This will set TERM to vt100
        # or sun if the terminal is --still-- undetected.
        # Otherwise sets LINES and COLUMNS env variables to response.
        self._check_naws()

        # this is totally useless, but informitive debugging information for
        # unknown unix clients ..
        # self._try_xtitle()
        # if not self.client.active:
        #    return

    def run(self):
        """
        Negotiate and inquire about terminal type, telnet options, window size,
        and tcp socket options before spawning a new session.
        """
        logger = logging.getLogger()
        from x84.bbs.exception import Disconnected
        try:
            self._set_socket_opts()
            self.banner()
            self._spawn_session()
        except socket.error as err:
            logger.debug('Connection closed: %s', err)
            self.client.deactivate()
        except Disconnected as err:
            logger.debug('Connection closed: %s', err)
            self.client.deactivate()

    def _timeleft(self, st_time):
        """
        Returns True when difference of current time and st_time is below
        TIME_WAIT_STAGE.
        """
        return bool(time.time() - st_time < self.TIME_WAIT_STAGE)

    def _set_socket_opts(self):
        """
        Set socket non-blocking and enable TCP KeepAlive.
        """
        self.client.sock.setblocking(0)
        self.client.sock.setsockopt(
            socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)

    def _try_env(self):
        """
        Try to snarf out some environment variables from unix machines.
        """
        if not self.client.active:
            return
        logger = logging.getLogger()
        from x84.telnet import NEW_ENVIRON, UNKNOWN
        # hard to tell if we already sent this once .. we mimmijammed
        # our own test ..
        if(self.client.ENV_REQUESTED and self.client.ENV_REPLIED):
            logger.debug('environment enabled (unsolicted)')
            return
        logger.debug('request-do-env')
        self.client.request_do_env()
        self.client.socket_send()  # push
        st_time = time.time()
        while (self.client.check_remote_option(NEW_ENVIRON) is UNKNOWN
                and not self.client.ENV_REPLIED
                and self._timeleft(st_time)
               and self.client.active):
            time.sleep(self.TIME_POLL)
        if not self.client.active:
            return
        if self.client.check_remote_option(NEW_ENVIRON) is UNKNOWN:
            logger.debug('failed: NEW_ENVIRON')

    def _check_naws(self):
        """
        Negotiate about window size (NAWS) telnet option (on).
        """
        if not self.client.active:
            return
        logger = logging.getLogger()
        if (self.client.env.get('LINES', None) is not None
                and self.client.env.get('COLUMNS', None) is not None):
            logger.debug('window size: %sx%s (unsolicited)',
                         self.client.env.get('COLUMNS'),
                         self.client.env.get('LINES'),)
            return
        st_time = time.time()
        while (self.client.env.get('LINES', None) is None
                and self.client.env.get('COLUMNS', None) is None
                and self._timeleft(st_time)
                and self.client.active):
            time.sleep(self.TIME_POLL)
        if (self.client.env.get('LINES', None) is not None
                and self.client.env.get('COLUMNS', None) is not None):
            logger.info('window size: %sx%s (negotiated)',
                        self.client.env.get('COLUMNS'),
                        self.client.env.get('LINES'))
            return
        if not self.client.active:
            return
        logger.debug('failed: negotiate about window size')
        self._try_cornerquery()

    def _try_xtitle(self):
        """
        request xterm title and store as _xtitle 'env' variable,
        this is restored on goodbye/logoff..

        actually, going to use restore codes in addition, to, nice to know
        anyway. may help with terminal id?
        """
        # P s = 2 1 -> Report xterm window's title. Result is OSC l label ST
        # http://invisible-island.net/xterm/ctlseqs/ctlseqs.html#VT100%20Mode
        # http://www.xfree86.org/4.5.0/ctlseqs.html#VT100%20Mode
        logger = logging.getLogger()
        logger.debug('report-xterm-title')
        self.client.send_str(chr(27) + '[21t')
        self.client.socket_send()  # push
        # response is '\x1b]lbash\x1b\\'
        response_pattern = re.compile(''.join((
            re.escape(chr(27)),
            r"\]l(.*)",
            re.escape(chr(27)),
            re.escape('\\'),)))
        st_time = time.time()
        while (self.client.idle() < self.TIME_WAIT_SILENT
               and self._timeleft(st_time)
               and self.client.active):
            time.sleep(self.TIME_POLL)
        if not self.client.active:
            return
        inp = self.client.get_input()
        match = response_pattern.search(inp)
        if not match:
            logger.debug('failed: xterm-title')
            return
        self.client.env['_xtitle'] = match.group(1).decode(
            'utf8', 'replace')
        logger.info('window title: %s', self.client.env['_xtitle'])
        self.client.send_str(chr(27) + '[20t')
        self.client.socket_send()  # push
        # response is '\x1b]Lbash\x1b\\'
        response_pattern = re.compile(''.join((
            re.escape(chr(27)),
            r"\]L(.*)",
            re.escape(chr(27)),
            re.escape('\\'),)))
        st_time = time.time()
        while (self.client.idle() < self.TIME_WAIT_SILENT * 2
               and self._timeleft(st_time)
               and self.client.active):
            time.sleep(self.TIME_POLL)
        if not self.client.active:
            return
        inp = self.client.get_input()
        match = response_pattern.search(inp)
        if not match:
            logger.debug('failed: xterm-icon')
            return
        self.client.env['_xicon'] = match.group(1).decode(
            'utf8', 'replace')
        logger.info('window icon: %s', self.client.env['_xicon'])

    def _try_cornerquery(self):
        """
        This is akin to X11's 'resize', move the cursor to the corner of the
        terminal (999,999) and request the terminal to report their cursor
        position.
        """
        logger = logging.getLogger()
        # Try #2 ... this works for most any screen
        # send to client --> pos(999,999)
        # send to client --> report cursor position
        # read from client <-- window size
        # bonus: 'vt100' or 'sun' TERM type set, lol.
        logger.debug('store-cu')
        self.client.send_str('\x1b[s')
        for kind, query_seq, response_pattern in self.WINSIZE_TRICK:
            logger.debug('move-to corner & query for %s', kind)
            # crashes syncterm with 999; changed to 255.
            self.client.send_str('\x1b[255;255')
            self.client.send_str(query_seq)
            self.client.socket_send()  # push
            st_time = time.time()
            while (self.client.idle() < self.TIME_WAIT_SILENT
                   and self._timeleft(st_time)
                   and self.client.active):
                time.sleep(self.TIME_POLL)
            if not self.client.active:
                return
            inp = self.client.get_input()
            self.client.send_str('\x1b[r')
            logger.debug('cursor restored')
            self.client.socket_send()  # push
            match = response_pattern.search(inp)
            if match:
                height, width = match.groups()
                self.client.rows = int(height)
                self.client.columns = int(width)
                logger.info('window size: %dx%d (corner-query hack)',
                            self.client.columns, self.client.rows)
                if self.client.env['TERM'] == 'unknown':
                    logger.warn("env['TERM'] = %r by POS", kind)
                    self.client.env['TERM'] = kind
                self.client.env['LINES'] = height
                self.client.env['COLUMNS'] = width
                return

        logger.debug('failed: negotiate about window size')
        # set to 80x24 if not detected
        self.client.columns, self.client.rows = 80, 24
        logger.debug('window size: %dx%d (default)',
                     self.client.columns, self.client.rows)
        self.client.env['LINES'] = str(self.client.rows)
        self.client.env['COLUMNS'] = str(self.client.columns)

    def _try_ttype(self):
        """
        Negotiate terminal type (TTYPE) telnet option (on).
        """
        if not self.client.active:
            return
        logger = logging.getLogger()
        detected = lambda: self.client.env['TERM'] != 'unknown'
        if detected():
            logger.debug('terminal type: %s (unsolicited)' %
                         (self.client.env['TERM'],))
            return
        logger.debug('request-terminal-type')
        st_time = time.time()
        if not detected():
            self.client.request_ttype()
            self.client.socket_send()  # push
        while (not detected() and self._timeleft(st_time)
               and self.client.active):
            time.sleep(self.TIME_POLL)
        if detected():
            logger.debug('terminal type: %s (negotiated by ttype)' %
                         (self.client.env['TERM'],))
            return
        if not self.client.active:
            return
        logger.warn('%r TERM not determined by TTYPE.', self.client.addrport())

    def _try_device_attributes(self):
        """
        Try to identify the type of DEC terminal. reports.c of vttest
        includes all the valid responses!
        """
        if not self.client.active:
            return
        logger = logging.getLogger()
        logger.debug('request-device-attributes')
        query_seq = bytes('\x1b[0c')
        self.client.send_str(query_seq)
        self.client.socket_send()  # push
        st_time = time.time()
        while (self._timeleft(st_time) and self.client.active):
            time.sleep(self.TIME_POLL)
        if not self.client.active:
            return
        inp = self.client.get_input()
        matched = False
        for p in range(len(inp)):
            for pattern, ttype, attrs in self.DA_REPLIES:
                if inp[p:].startswith(pattern):
                    matched = (inp[p:p+len(pattern)], ttype, attrs)
                    break
            if matched:
                break
        if matched:
            sequence, ttype, attrs = matched
            if len(sequence) != len(inp):
                # if we received more bytes than just the sequence, we
                # don't really care to rewind and buffer the rest for
                # later interpretation.
                toss_left = inp[:inp.find(sequence)] 
                toss_right = inp[inp.find(sequence) + len(sequence):]
                logger.warn("threw out during da reply: %r, <DA>, %r.",
                        toss_left, toss_right)
                self.client.env['DA'] = matched
            if (self.client.env.get('TERM', 'unknown')
                    in ('unknown', 'vt100', 'sun')):
                    logger.debug("%r terminal type: %s by da",
                            self.client.addrport(), ttype)
                    self.client.env['TERM'] = ttype
                    # custom environment variable, we worked so hard to get it,
                    self.client.env['TERM_ATTRS'] = attrs  # why throw it away?
            else:
                logger.info('%r adheres to %s;%s',
                        self.client.addrport(), ttype, attrs)
                self.client.env['TERM_ALT'] = ttype
                self.client.env['TERM_ATTRS'] = attrs
        else:
            logger.warn('%r no DEC VT response.', self.client.addrport())
            return




"""
bbs lister for x/84, http://github.com/jquast/x84

  To use the (optional) http://bbs-scene.org API,
  configure a section in your .ini file:

    [bbs-scene]
    user = my@email-addr.ess
    pass = my-plaintext-password
"""
import xml.etree.ElementTree
import threading
import requests
import logging
import time
import math
import os

DB_KEYS = ('bbsname', 'sysop', 'software',
           'address', 'port', 'location',
           'notes', 'timestamp', 'ansi')
XML_REQNOTNULL = ('bbsname', 'sysop', 'software',
                  'address', 'location', 'notes')

import sauce


def fancy_blue(char, blurb=u''):
    """ Diplay '(char)blurb' in fancy blue. """
    from x84.bbs import getterminal
    term = getterminal()
    return u''.join((
        term.bold_blue('('),
        term.blue_reverse(char),
        term.bold_blue(')'),
        term.bold_white(blurb),))


def fancy_green(char, blurb=u''):
    """ Diplay '(char)blurb' in fancy green. """
    from x84.bbs import getterminal
    term = getterminal()
    return u''.join((
        term.bold_green('('),
        term.green_reverse(char),
        term.bold_green(')'),
        term.bold_white(blurb),))


class FetchUpdates(threading.Thread):
    """ Fetch bbs-scene.org bbs list as thread. """
    url = 'http://bbs-scene.org/api/bbslist.php'
    content = list()

    def run(self):
        """ Begin fetch, results are stored in self.content """
        from x84.bbs import ini
        logger = logging.getLogger()
        usernm = ini.CFG.get('bbs-scene', 'user')
        passwd = ini.CFG.get('bbs-scene', 'pass')
        logger.info('fetching %s ..', self.url)
        # pylint: disable=E1103
        req = requests.get(self.url, auth=(usernm, passwd))
        if 200 != req.status_code:
            logger.error(req.content)
            logger.error('bbs-scene.org: %s', req.status_code)
            return
        logger.info('bbs-scene.org: %d', req.status_code)

        # some byte-level pre-parsing sanitization
        buf = ''.join((byte for byte in req.content
                       if ord(byte) >= 0x20
                       or ord(byte) in (0x09, 0x0a, 0x0d)))

        # append elements into self.content (bbs_id, dict(attributes))
        for node in xml.etree.ElementTree.XML(buf).findall('node'):
            bbs_id = node.find('id').text.strip()
            record = dict([(key,
                ((node.find(key).text or u'').strip()
                    if node.find(key) is not None else u''))
                for key in DB_KEYS])
            self.content.append((bbs_id, record))


def wait_for(thread):
    """ for dummy pager, wait for return value before listing, """
    from x84.bbs import echo, getch
    wait_fetch = 8
    if thread.is_alive():
        echo(u"\r\n\r\nfEtchiNG bbS-SCENE.ORG bbS liSt.. "
             u"(%s)s\b\b%s" % (' ' * 2, '\b' * 2,))
        for num in range(wait_fetch):
            echo('%2d%s' % (wait_fetch - num - 1, u'\b' * 2,))
            if not thread.is_alive():
                return
            thread.join(1)
            if getch(0) == u'q':
                # undocumented: q cancels -- in case it ever goes down D:
                return


def chk_thread(thread):
    """
    for asyncronous lightbar, return False if thread is still running,
    'dirty' if thread completed and new entries were found, 'None' if thread
    completed and no no entries were discovered.
    """
    from x84.bbs import getsession, DBProxy
    if thread is None or thread.is_alive():
        return False

    # check if bbs-scene.org thread finished, silly hack
    logger = logging.getLogger()
    session = getsession()
    udbkeys = DBProxy('bbslist').keys()
    nlc = 0
    for key, value in thread.content:
        if key not in udbkeys:
            DBProxy('bbslist')[key] = value
            DBProxy('bbslist', 'comments')[key] = list()
            DBProxy('bbslist', 'ratings')[key] = list()
            nlc += 1
        else:
            # update anyway (fe. port changed), we don't
            # detect this as an update.
            DBProxy('bbslist')[key] = value
    if nlc > 0:
        logger.info('%d new entries', nlc)
        session.send_event('global', ('bbslist_update', None))
        session.buffer_event('bbslist_update')
        return 'dirty'
    else:
        logger.info('no new bbs-scene.org entries')
    return None


def get_bbslist(max_len=23):
    """
    Returns tuple, (bbs_key, display_string), grouped by bbs software !
    """
    from x84.bbs import getsession, DBProxy, getterminal
    session, term = getsession(), getterminal()
    session.flush_event('bbslist_update')

    def get_bysoftware():
        """
        Group bbs list by software
        """
        by_group = dict()
        for (key, bbs) in DBProxy('bbslist').iteritems():
            grp = bbs['software'].decode('iso8859-1', 'replace')
            if 'citadel' in grp.lower():
                grp = u'Citadel'
            elif 'mystic' in grp.lower():
                grp = u'Mystic'
            elif 'synchronet' in grp.lower():
                grp = u'Synchronet'
            if not grp.lower().startswith('the'):
                grp = grp.split()[0].title()
            else:
                grp = grp.title().split()[1]
            if not grp in by_group:
                by_group[grp] = [(key, bbs)]
                continue
            by_group[grp].append((key, bbs))
        return by_group.items()
    def strip_name(key):
        if len(key) >= max_len:
            key = key[:(max_len - 2)] + ' $'
        return key
    output = list()
    for _grp, _value in sorted(get_bysoftware()):
        output.append((None, term.blue(_grp.rjust(max_len))))
        for key, bbs in sorted(_value):
            output.append((key, strip_name(bbs['bbsname'])))
    return output


def view_ansi(key):
    """ fetch and view a bbs ansi. They're not often very good ...
    """
    from x84.bbs import getterminal, echo, DBProxy, ini, getch, from_cp437
    term = getterminal()
    ansiurl = DBProxy('bbslist')[key]['ansi']
    logger = logging.getLogger()
    echo(u'\r\n\r\n')
    if ansiurl is not None and 0 != len(ansiurl) and ansiurl != 'NONE':
        usernm = ini.CFG.get('bbs-scene', 'user')
        passwd = ini.CFG.get('bbs-scene', 'pass')
        req = requests.get(ansiurl, auth=(usernm, passwd))
        if req.status_code != 200:
            echo(u'\r\n\r\nrequest failed,\r\n')
            echo(u'%r' % (req.content,))
            echo(u'\r\n\r\n(code : %s).\r\n' % (req.status_code,))
            echo(u'\r\nPress any key ..')
            logger.warn('ansiurl request failed: %s' % (ansiurl,))
            getch()
            return
        ansi_txt = from_cp437(sauce.SAUCE(data=req.content).__str__())
        echo(ansi_txt)
    else:
        echo('no ansi available (%s)' % (ansiurl,))
    # move to bottom of screen and getch
    echo(u''.join((
        term.move(term.height, 0),
        term.normal,
        u'\r\n\r\nPRESS ANY kEY ...'),))
    getch()


def calc_rating(ratings, outof=4):
    """ Given set of ratings, return ucs \*, of 1 to *outouf*, star rating. """
    if 0 == len(ratings):
        return u'-'
    total = sum([_rating for (_handle, _rating) in ratings])
    stars = int(math.floor((total / len(ratings))))
    return u'%*s' % (outof, u'*' * stars)


def get_bbsinfo(key, active=True):
    """
    given a bbs key, fetch detailed information for use in pager
    """
    # pylint: disable=R0914
    #        Too many local variables.
    from x84.bbs import getterminal, DBProxy, timeago
    rstr = u''
    term = getterminal()
    highlight = term.bold_blue if active else term.blue
    lowlight = term.bold_black
    bbs = DBProxy('bbslist')[key]
    epoch = time.mktime(time.strptime(bbs['timestamp'], '%Y-%m-%d %H:%M:%S'))
    ratings = DBProxy('bbslist', 'ratings')[key]
    comments = DBProxy('bbslist', 'comments')[key]
    rstr += (lowlight('bbSNAME')
             + highlight(u': ')
             + bbs['bbsname']
             + highlight('  +o ')
             + bbs['sysop'] + u'\r\n')
    rstr += (lowlight('AddRESS')
             + highlight(': ')
             + bbs['address']
             + highlight(': ')
             + bbs['port'] + u'\r\n')
    rstr += (lowlight('lOCAtiON')
             + highlight(': ')
             + bbs['location'] + '\r\n')
    rstr += (lowlight('SOftWARE')
             + highlight(': ')
             + bbs['software'] + '\r\n')
    rstr += (lowlight('tiMEStAMP')
             + highlight(':')
             + lowlight(timeago(time.time() - epoch))
             + ' ago\r\n')
    rstr += (lowlight('RAtiNG') + highlight(': ')
             + '%s (%2.2f of %d)\r\n' % (
                 highlight(calc_rating(ratings)),
                 0 if 0 == len(ratings) else
                 sum([_rating for (_handle, _rating) in ratings])
                 / len(ratings), len(ratings)))
    rstr += u'\r\n' + bbs['notes']
    for handle, comment in comments:
        rstr += '\r\n\r\n' + lowlight(handle)
        rstr += highlight(': ')
        rstr += comment
    return rstr


def get_swinfo(entry, pager):
    """
    given a normalized bbs software name,
    fetch a description paragraph for use in pager
    """
    from x84.bbs import getterminal
    from x84.bbs.output import Ansi
    term = getterminal()
    output = pager.clear()
    if entry:
        entry = Ansi(entry).seqfill().strip()
    if entry and entry.strip().lower() == 'enthral':
        output += pager.update(
            "Enthral is a fresh look at the old school art of bbsing. "
            "It's a fresh face to an old favorite. Although Enthral is "
            "still in it's alpha stages, the system is quite stable and "
            "is already very feature rich. Currently available for "
            "Linux, BSD, and Apple's OS X.\r\n\r\n"
            "   " + term.bold_blue('http://enthralbbs.com/') + "\r\n\r\n"
            "Author: Mercyful Fate\n"
            "IRC: #enthral on irc.bbs-scene.org\r\n")
        output += pager.title(u'- about ' + term.blue('Enthral') + u' -')
    elif entry and entry.strip().lower() == 'citadel':
        output += pager.update(
            "Ancient history.\r\n\r\n")
        output += pager.title(u'- about ' + term.blue('Citadel') + u' -')
    elif entry and entry.strip().lower() == 'mystic':
        output += pager.update(
            "Mystic BBS is a bulletin board system (BBS) software in "
            "the vein of other \"forum hack\" style software such as "
            "Renegade, Oblivion/2, and Iniquity. Like many of its "
            "counterparts it features a high degree of relatively "
            "easy customization thanks to its ACS based menu system "
            "along with fully editable strings and ANSI themes. "
            "Mystic also includes its own Pascal like MPL scripting "
            "language for even further flexibility.\r\n\r\n"
            "  " + term.bold_blue('http://mysticbbs.com/') + "\r\n\r\n"
            "Author: g00r00\r\n"
            "IRC: #MysticBBS on irc.efnet.org\r\n")
        output += pager.title(u'- about ' + term.blue('Mystic') + u' -')
    elif entry and entry.strip().lower() == 'synchronet':
        output += pager.update(
            "Synchronet Bulletin Board System Software is a free "
            "software package that can turn your personal computer "
            "into your own custom online service supporting multiple "
            "simultaneous users with hierarchical message and file "
            "areas, multi-user chat, and the ever-popular BBS door "
            "games.\r\n\r\n"
            "Synchronet has since been substantially redesigned as "
            "an Internet-only BBS package for Win32 and Unix-x86 "
            "platforms and is an Open Source project under "
            "continuous development.\r\n\r\n"
            "  " + term.bold_blue('http://www.synchro.net/\r\n') + "\r\n\r\n"
            "Author: Deuce\r\n"
            "IRC: #synchronet on irc.bbs-scene.org")
        output += pager.title(u'- about ' + term.blue('Synchronet') + u' -')
    elif entry and entry.strip().lower() == 'progressive':
        output += pager.update(
            "This bbs features threading, intra-process communication, "
            "and easy scripting in python. X/84 is a continuation of "
            "this codebase.\r\n\r\n"
            + "Author: jojo\r\n"
            "IRC: #prsv on irc.efnet.org")
        output += pager.title(u'- about ' + term.blue('The Progressive -'))
    elif entry and entry.strip().lower() == 'x/84':
        output += pager.update(
            "X/84 is an open source python utf8 bsd-licensed telnet "
            "server specificly designed for BBS's, MUD's, and high "
            "scriptability. It is a Continuation of 'The Progressive' "
            "and is the only BBS software to support both CP437 and "
            "UTF8 encoding.\r\n\r\n"
            "  " + term.bold_blue('https://github.com/jquast/x84/\r\n')
            + "\r\n\r\nAuthor: dingo, jojo\r\n"
            "IRC: #prsv on irc.efnet.org")
        output += pager.title(u'- about ' + term.blue('X/84') + u' -')
    else:
        output += pager.update(u' no information about %s.'
                               % (entry or u'').title(),)
        output += pager.title(u'- about ' + term.blue(entry or u'') + u' -')
    return output


def get_ui(position=None):
    """
    Returns user interface (lightbar, pager).
    Optional argument position is tuple position of prior lightbar instance.
    """
    from x84.bbs import getterminal, Lightbar, Pager
    term = getterminal()
    assert term.height > 10 and term.width >= 40
    # +-+ +----+
    # |lb |pager
    # +-+ +----+
    height = term.height - 7
    lb_width = int(term.width * .3)
    pg_width = term.width - (lb_width)
    lb_xloc = (term.width / 2) - (term.width / 2)
    pg_xloc = lb_xloc + lb_width
    lightbar = Lightbar(height, lb_width, (term.height - height - 1), lb_xloc)
    pager = Pager(height, pg_width, (term.height - height - 1), pg_xloc)
    pager.ypadding = 2
    pager.xpadding = 2
    lightbar.update(get_bbslist(max_len=lightbar.visible_width))
    ## pressing Return is same as 't'elnet
    lightbar.keyset['enter'].extend((u't', u'T'))
    ## re-select previous selection
    if position is not None:
        lightbar.position = position
    return (pager, lightbar)


def banner():
    """ Display banner/art. """
    from x84.bbs import getterminal
    term = getterminal()
    artfile = os.path.join(os.path.dirname(__file__), 'art', 'bbslist.asc')
    return u''.join((
        u'\r\n\r\n',
        u'\r\n'.join([line.center(term.width)[:term.width].rstrip()
                      for line in open(artfile)]) if os.path.exists(artfile)
        else u'',))


def redraw_pager(pager, selection, active=True):
    """
    Display bbs information in pager window for given selection.
    """
    from x84.bbs import getterminal, DBProxy
    term = getterminal()
    output = u''
    pager.colors['border'] = term.blue if active else term.bold_black
    output += pager.border()
    output += pager.clear()
    pager.move_home()
    key, entry = selection
    if key is None:
        output += get_swinfo((entry or u'?').strip(), pager)
        output += pager.footer(u'-'
                               + fancy_green('a', 'dd') + u' -')
    else:
        # describe bbs in pager
        bbsname = DBProxy('bbslist')[key]['bbsname']
        ansiurl = DBProxy('bbslist')[key].get('ansi', 'NONE')
        output += pager.title(u'- %s -' % (
            term.bold_blue(bbsname) if active else term.blue(bbsname)))
        output += pager.update(get_bbsinfo(key, active))
        ans = u''
        if ansiurl != 'NONE' and 0 != len(ansiurl):
            ans = u'.' + fancy_green('v', 'ansi')
        output += pager.footer(u'- '
                               + fancy_green('a', 'dd') + u'.'
                               + fancy_green('t', 'ElNEt') + u'.'
                               + fancy_green('c', 'OMNt') + u'.'
                               + fancy_green('r', 'AtE') + ans + u' -')
        output += pager.pos(pager.height - 2, pager.width - 4)
    # pager scroll indicator ..
    output += pager.pos(pager.height - 2, pager.width - 4)
    if pager.bottom != pager.position:
        output += fancy_green('+') if active else fancy_blue('+')
    else:
        output += '   '
    return output


def redraw_lightbar(lightbar, active=True):
    """
    Display bbs listing in lightbar.
    """
    from x84.bbs import getterminal
    term = getterminal()
    lightbar.colors['border'] = term.bold_green if active else term.bold_black
    output = lightbar.border()
    if active:
        lightbar.colors['selected'] = term.green_reverse
    else:
        lightbar.colors['selected'] = term.blue_reverse
#        output += lightbar.footer(u''.join((
#            u'- ',
#            fancy_green('up', '.'),
#            fancy_green('down', '.'),
#            fancy_green('right', u''),
#            u' -')))
#        output += lightbar.footer(u''.join((
#            u'- ',
#            fancy_blue('up', '.'),
#            fancy_blue('down', '.'),
#            fancy_green('left', u''),
#            u' -')))
    output += lightbar.title(u'- ' + fancy_green('a', 'add') + ' -')
    output += lightbar.refresh()
    return output


def redraw(pager, lightbar, leftright):
    """
    Redraw pager and lightbar.
    """
    return (redraw_pager(pager,
                         selection=lightbar.selection,
                         active=(leftright == 1))
            + redraw_lightbar(lightbar, active=(leftright == 0)))


def more(cont=False):
    """
    Returns True if user 'q'uit; otherwise False
    when prompting is complete (moar/next/whatever)
    """
    from x84.bbs import echo, getch, Ansi, getterminal, LineEditor, DBProxy
    prompt_key = u'\r\n\r\nENtER BBS iD: '
    msg_badkey = u'\r\n\r\nbbS id iNVAliD!'
    term = getterminal()
    prompt = u', '.join(fancy_blue(char, blurb)
                        for char, blurb in
                        (('i', 'NfO',),
                         ('a', 'dd',),
                         ('c', 'OMMENt',),
                         ('r', 'AtE',),
                         ('t', 'ElNEt',),
                         ('v', 'ANSi'),
                         ('q', 'Uit',)))
    if cont:
        prompt += u', ' + fancy_blue(' ', 'more')
    prompt += u': '
    while True:
        echo('\r\n' + Ansi(prompt).wrap(term.width - (term.width / 3)))
        inp = getch()
        if inp in (u'q', 'Q'):
            return True
        elif inp is not None and type(inp) is not int:
            if cont and inp == u' ':
                echo('\r\n\r\n')
                return False
            if inp.lower() in u'acrtviACRTVI':
                # these keystrokes require a bbs key argument,
                # prompt the user for one
                echo(prompt_key)
                key = LineEditor(5).read()
                if (key is None or 0 == len(key.strip())
                        or not key in DBProxy('bbslist')):
                    echo(msg_badkey)
                    continue
                process_keystroke(inp, key)


def dummy_pager():
    """
    Provide interface without pager/lightbar.
    """
    # pylint: disable=R0912
    #        Too many branches
    from x84.bbs import getterminal, echo, getch, Ansi
    term = getterminal()
    msg_header = u'// bbS liSt'
    hindent = 2
    vindent = 5
    nlines = 0
    bbslist = get_bbslist()
    echo(u'\r\n' + msg_header.center(term.width).rstrip() + '\r\n\r\n')
    if 0 == len(bbslist):
        echo(u'\r\n\r\nNO BBSS. a%sdd ONE, q%sUit' % (
            term.bold_blue(':'), term.bold_blue(':')))
        while True:
            inp = getch()
            if inp in (u'q', 'Q'):
                return  # quit
            elif inp in (u'a', 'A'):
                process_keystroke(inp)
                break
    while True:
        for (key, line) in bbslist:
            if key is None:  # bbs software
                echo(term.blue_reverse(line.rstrip()) + '\r\n')
                nlines += 1
            else:
                wrapd = Ansi(line).wrap(term.width - hindent)
                echo(term.bold_blue(key) + term.bold_black('. '))
                for num, line in enumerate(wrapd.split('\r\n')):
                    if num != 0:
                        echo(' ' * hindent)
                    echo(line + '\r\n')
                    nlines += 1
            if nlines and (nlines % (term.height - vindent) == 0):
                if more(True):
                    return
        # one final prompt before exit
        if more(False):
            return
    return


def add_bbs():
    """
    Prompt user for details and to add bbs to list.
    """
    # pylint: disable=R0914,R0915
    #        Too many local variables
    #        Too many statements
    from x84.bbs import getsession, getterminal, echo, LineEditor, DBProxy, ini
    from x84.bbs import getch
    session, term = getsession(), getterminal()
    echo(term.move(term.height, 0))
    empty_msg = u'\r\n\r\nVAlUE iS NOt OPtiONAl.'
    cancel_msg = u"\r\n\r\nENtER 'quit' tO CANCEl."
    saved_msg = u'\r\n\r\nSAVED AS RECORd id %s.'
    logger = logging.getLogger()
    bbs = dict()
    for bkey in DB_KEYS:
        if bkey == 'timestamp':
            value = time.strftime('%Y-%m-%d %H:%M:%S')
            bbs[bkey] = value
            continue
        elif bkey == 'ansi':
            # todo: upload ansi with xmodem .. lol !?
            value = u''
            continue
        splice = len(bkey) - (len(bkey) / 3)
        prefix = (u'\r\n\r\n  '
                  + (term.bold_red('* ') if bkey in XML_REQNOTNULL else u'')
                  + term.bold_blue(bkey[:splice])
                  + term.bold_black(bkey[splice:])
                  + term.bold_white(': '))
        led = LineEditor(40)  # !?
        led.highlight = term.green_reverse
        while True:
            echo(prefix)
            value = led.read()
            if value is not None and (value.strip().lower() == 'quit'):
                return
            if bkey in XML_REQNOTNULL:
                if value is None or 0 == len(value.strip()):
                    echo(term.bold_red(empty_msg))
                    echo(u'\r\n' + cancel_msg)
                    continue
            if bkey in ('port') and value is None or 0 == len(value):
                value = u'23'
            # TODO: telnet connect test, of course !
            bbs[bkey] = value
            break
    key = max([int(_key) for _key in DBProxy('bbslist').keys()] or [0]) + 1
    DBProxy('bbslist')[key] = bbs
    DBProxy('bbslist', 'comments')[key] = list()
    DBProxy('bbslist', 'ratings')[key] = list()
    echo('\r\n\r\n' + saved_msg % (key) + '\r\n')
    session.send_event('global', ('bbslist_update', None,))
    session.buffer_event('bbslist_update')
    if ini.CFG.has_section('bbs-scene'):
        # post to bbs-scene.org
        posturl = 'http://bbs-scene.org/api/bbslist.xml'
        usernm = ini.CFG.get('bbs-scene', 'user')
        passwd = ini.CFG.get('bbs-scene', 'pass')
        data = {'name': bbs['bbsname'],
                'sysop': bbs['sysop'],
                'software': bbs['software'],
                'address': bbs['address'],
                'port': bbs['port'],
                'location': bbs['location'],
                'notes': bbs['notes'],
                }
        req = requests.post(posturl, auth=(usernm, passwd), data=data)
        if req.status_code != 200:
            echo(u'\r\n\r\nrequest failed,\r\n')
            echo(u'%r' % (req.content,))
            echo(u'\r\n\r\n(code : %s).\r\n' % (req.status_code,))
            echo(u'\r\nPress any key ..')
            logger.warn('bbs post failed: %s' % (posturl,))
            getch()
            return
        logger.info('bbs-scene.org api (%d): %r/%r',
                    req.status_code, session.user.handle, bbs)
        # spawn a thread to re-fetch bbs entries,
        thread = FetchUpdates()
        thread.start()
        wait_for(thread)
        return chk_thread(thread)


def process_keystroke(inp, key=None):
    """ Process general keystroke and call routines.  """
    from x84.bbs import getsession, DBProxy, gosub, echo
    session = getsession()
    if inp is None:
        return False
    elif type(inp) is int:
        return False
    elif inp in (u't', u'T') and key is not None:
        bbs = DBProxy('bbslist')[key]
        gosub('telnet', bbs['address'], bbs['port'],)
    elif inp in (u'a', u'A'):
        add_bbs()
    elif inp in (u'c', u'C') and key is not None:
        add_comment(key)
    elif inp in (u'r', u'R') and key is not None:
        rate_bbs(key)
    elif inp in (u'i', u'I') and key is not None:
        echo(get_bbsinfo(key) + '\r\n')
    elif inp in (u'v', u'V') and key is not None:
        view_ansi(key)
    elif inp in (u'd', u'D') and key is not None and (
            'sysop' in session.user.groups):
        del DBProxy('bbslist')[key]
    else:
        return False  # unhandled
    return True


def add_comment(key):
    """ Prompt user to add a comment about a bbs. """
    # pylint: disable=R0914
    #        Too many local variables.
    from x84.bbs import getsession, getterminal, echo, DBProxy, LineEditor
    from x84.bbs import getch
    session, term = getsession(), getterminal()
    prompt_comment = u'\r\n\r\nWhAt YOU GOt tO SAY? '
    prompt_chg = u'\r\n\r\nChANGE EXiStiNG ? [yn] '
    echo(term.move(term.height, 0))
    echo(prompt_comment)
    comment = LineEditor(max(10, term.width - len(prompt_comment) - 5)).read()
    if comment is None or 0 == len(comment.strip()):
        return
    new_entry = (session.handle, comment)
    comments = DBProxy('bbslist', 'comments')
    comments.acquire()
    existing = comments[key]
    if session.handle in (_nick for (_nick, _cmt) in comments[key]):
        # change existing comment,
        echo(prompt_chg)
        if getch() not in (u'y', u'Y'):
            comments.release()
            return
        # re-define list without existing entry, + new entry
        comments[key] = [(_enick, _ecmt) for (_enick, _ecmt) in existing
                         if session.handle != _enick] + [new_entry]
        comments.release()
        return
    # re-define as existing list + new entry
    comments[key] = existing + [new_entry]
    comments.release()


def rate_bbs(key):
    """ Prompt user to rate a bbs. """
    # pylint: disable=R0914
    #        Too many local variables
    from x84.bbs import getsession, getterminal, echo, LineEditor, DBProxy
    from x84.bbs import getch
    session, term = getsession(), getterminal()
    prompt_rating = u'\r\n\r\nRAtE 0.0 - 4.0: '
    prompt_chg = u'\r\n\r\nChANGE EXiStiNG ? [yn] '
    msg_invalid = u'\r\n\r\niNVAlid ENtRY.\r\n'
    echo(term.move(term.height, 0) + '\r\n')
    echo(prompt_rating)
    s_rating = LineEditor(3).read()
    if s_rating is None or 0 == len(s_rating.strip()):
        return
    try:
        f_rating = float(s_rating)
    except ValueError:
        echo(msg_invalid)
        return
    if f_rating < 0 or f_rating > 4:
        echo(msg_invalid)
        return
    entry = (session.handle, f_rating)
    ratings = DBProxy('bbslist', 'ratings')
    ratings.acquire()
    if session.handle in (_handle for (_handle, _rating) in ratings[key]):
        echo(prompt_chg)
        if getch() not in (u'y', u'Y'):
            ratings.release()
            return
        # re-define list without existing entry, + new entry
        ratings[key] = [(__handle, __rating)
                        for (__handle, __rating) in ratings[key]
                        if session.handle != __handle] + [entry]
        ratings.release()
        return
    # re-define as existing list + new entry
    ratings[key] = ratings[key] + [entry]
    ratings.release()


def main():
    """ Main procedure. """
    # pylint: disable=R0914,R0912,R0915
    #        Too many local variables
    #        Too many branches
    #        Too many statements
    from x84.bbs import getsession, getterminal, echo, ini, getch, gosub
    from x84.bbs import DBProxy
    session, term = getsession(), getterminal()
    pager, lightbar = get_ui(None)

    echo(u'\r\n\r\n')
    thread = None
    if ini.CFG.has_section('bbs-scene'):
        thread = FetchUpdates()
        thread.start()
        session.activity = u'bbs lister [bbs-scene.org]'
        echo(u'fetching bbs-scene.org updates ...')
    else:
        session.activity = u'bbs lister'

    dirty = True
    # session.buffer_event('refresh', ('init',))
    leftright = 0  # 'left'

    while True:
        # check if screen requires refresh of any kind,
        if session.poll_event('refresh'):
            dirty = True
        if thread is not None:
            t_val = chk_thread(thread)
            if t_val != False:
                thread = None
                if t_val == 'dirty':
                    dirty = True
        if session.poll_event('bbslist_update'):
            while session.read_event('bbslist_update', 0.15):
                # pylint: disable=W0104
                #         Statement seems to have no effect
                None
            dirty = True

        # refresh advanced screen with lightbar and pager
        if dirty:
            if session.user.get('expert', False):
                # provide dumb terminal with hotkey prompt
                if thread is not None:
                    wait_for(thread)
                if chk_thread(thread) != False:
                    thread = None
                return dummy_pager()
            else:
                pager, lightbar = get_ui(lightbar.position)
                echo(banner())
                echo(u'\r\n' * lightbar.height)
                echo(redraw(pager, lightbar, leftright))
                dirty = False

        # detect and process keyboard input for advanced screen
        inp = getch(1)
        if inp == term.KEY_LEFT:
            # full refresh for border chang ;/
            leftright = 0
            echo(redraw_pager(
                pager, lightbar.selection, active=(leftright == 1)))
            echo(redraw_lightbar(
                lightbar, active=(leftright == 0)))
        elif inp == term.KEY_RIGHT:
            # full refresh for border chang ;/
            leftright = 1
            echo(redraw_pager(pager,
                              lightbar.selection, active=(leftright == 1)))
            echo(redraw_lightbar(lightbar,
                                 active=(leftright == 0)))
        elif inp is not None:
            # process as pager or lightbar keystroke,
            echo(lightbar.process_keystroke(inp)
                 if leftright == 0 else
                 pager.process_keystroke(inp))

            # full refresh after rate/comment/elnet/view etc.
            if process_keystroke(inp, lightbar.selection[0]):
                session.buffer_event('refresh', ('redraw',))
                continue

            # quit 'q'
            if (lightbar.quit or pager.quit):
                return

            # pressed return, telnet!
            if lightbar.selected and lightbar.selection[0] is not None:
                bbs = DBProxy('bbslist')[lightbar.selection[0]]
                gosub('telnet', bbs['address'], bbs['port'])
                # buffer full refresh after telnet
                session.buffer_event('refresh', ('redraw',))

            # selected new entry, refresh entire pager, a little bit
            # bandwidth excessive as bbs name is part of border title.
            if lightbar.moved:
                echo(redraw_pager(pager, lightbar.selection,
                                  active=(leftright == 1)))
                lightbar.moved = False



""" Chat script for x/84, https://github.com/jquast/x84 """

# unfortunately the hardlinemode stuff that irssi and such uses
# is not used, so each line causes a full screen fresh ..

import time
POLL_KEY = 0.15  # blocking ;; how often to poll keyboard
POLL_OUT = 0.25  # seconds elapsed before screen update, prevents flood

CHANNEL = None
NICKS = dict()
EXIT = False


def show_help():
    """ return string suitable for response to /help. """
    return u'\n'.join((
        u'   /join #channel',
        u'   /act mesg',
        u'   /part [reason]',
        u'   /quit [reason]',
        u'   /users',
        u'   /whois handle',))


def process(mesg):
    """
    Process a command recieved by event system. and return string
    suitable for displaying in chat window.
    """
    from x84.bbs import getsession
    session = getsession()
    sid, tgt_channel, (handle, cmd, args) = mesg
    ucs = u''
    # pylint: disable=W0602
    # Using global for 'NICKS' but no assignment is done
    global NICKS
    if (CHANNEL != tgt_channel and 'sysop' not in session.user.groups):
        ucs = u''
    elif cmd == 'join':
        if handle not in NICKS:
            NICKS[handle] = sid
            ucs = show_join(handle, sid, tgt_channel)
    elif handle not in NICKS:
        NICKS[handle] = sid
    elif cmd == 'part':
        if handle in NICKS:
            del NICKS[handle]
        ucs = show_part(handle, sid, tgt_channel, args)
    elif cmd == 'say':
        ucs = show_say(handle, tgt_channel, args)
    elif cmd == 'act':
        ucs = show_act(handle, tgt_channel, args)
    else:
        ucs = u'unhandled: %r' % (mesg,)
    return ucs


def show_act(handle, tgt_channel, mesg):
    """ return terminal sequence for /act performed by handle. """
    from x84.bbs import getsession, getterminal
    session, term = getsession(), getterminal()
    return u''.join((
        time.strftime('%H:%M'), u' * ',
        (term.bold_green(handle) if handle != session.handle
            else term.green(handle)),
        (u':%s' % (tgt_channel,)
            if 'sysop' in session.user.groups
            else u''), u' ',
        mesg,))


def show_join(handle, sid, chan):
    """ return terminal sequence for /join performed by handle. """
    from x84.bbs import getsession, getterminal
    session, term = getsession(), getterminal()
    return u''.join((
        time.strftime('%H:%M'), u' ',
        term.blue('-'), u'!', term.blue('-'),
        u' ', term.bold_cyan(handle), u' ',
        (u''.join((term.bold_black('['),
                   term.cyan(sid), term.bold_black(']'), u' ',))
            if 'sysop' in session.user.groups else u''),
        'has joined ',
        term.bold(chan),))


def show_part(handle, sid, chan, reason):
    """ return terminal sequence for /part performed by handle. """
    from x84.bbs import getsession, getterminal
    session, term = getsession(), getterminal()
    return u''.join((
        time.strftime('%H:%M'), u' ',
        term.blue('-'), u'!', term.blue('-'),
        u' ', term.bold_cyan(handle), u' ',
        (u''.join((term.bold_black('['),
                   term.cyan(sid), term.bold_black(']'), u' ',))
            if 'sysop' in session.user.groups else u''),
        'has left ',
        term.bold(chan),
        u' (%s)' % (reason,) if reason and 0 != len(reason) else u'',))


def show_whois(attrs):
    """ return terminal sequence for /whois result. """
    from x84.bbs import getsession, getterminal, timeago
    session, term = getsession(), getterminal()
    return u''.join((
        time.strftime('%H:%M'), u' ',
        term.blue('-'), u'!', term.blue('-'),
        u' ', term.bold(attrs['handle']), u' ',
        (u''.join((term.bold_black('['),
                   term.cyan(attrs['sid']), term.bold_black(']'), u' ',))
            if 'sysop' in session.user.groups else u''), u'\n',
        time.strftime('%H:%M'), u' ',
        term.blue('-'), u'!', term.blue('-'),
        u' ', u'CONNECtED ',
        term.bold_cyan(timeago(time.time() - attrs['connect_time'])),
        ' AGO.', u'\n',
        time.strftime('%H:%M'), u' ',
        term.blue('-'), u'!', term.blue('-'),
        u' ', term.bold(u'idlE: '),
        term.bold_cyan(timeago(time.time() - attrs['idle'])), u'\n',
    ))


def show_nicks(handles):
    """ return terminal sequence for /users result. """
    from x84.bbs import getterminal
    term = getterminal()
    return u''.join((
        time.strftime('%H:%M'), u' ',
        term.blue('-'), u'!', term.blue('-'),
        u' ', term.bold_cyan('%d' % (len(handles))), u' ',
        u'user%s: ' % (u's' if len(handles) > 1 else u''),
        u', '.join(handles) + u'\n',))


def show_say(handle, tgt_channel, mesg):
    """ return terminal sequence for /say performed by handle. """
    from x84.bbs import getsession, getterminal, get_user
    session, term = getsession(), getterminal()
    return u''.join((
        time.strftime('%H:%M'), u' ',
        term.bold_black(u'<'),
        (term.bold_red(u'@') if handle != 'anonymous'
            and 'sysop' in get_user(handle).groups
            else u''),
        (handle if handle != session.handle
            else term.bold(handle)),
        (u':%s' % (tgt_channel,)
            if 'sysop' in session.user.groups
            else u''),
        term.bold_black(u'>'), u' ',
        mesg,))


def get_inputbar(pager):
    """ Return ScrollingEditor for use as inputbar. """
    from x84.bbs import getterminal, ScrollingEditor
    term = getterminal()
    width = pager.visible_width - 2
    yloc = (pager.yloc + pager.height) - 2
    xloc = pager.xloc + 2
    ibar = ScrollingEditor(width, yloc, xloc)
    ibar.enable_scrolling = True
    ibar.max_length = 512
    ibar.colors['highlight'] = term.cyan_reverse
    return ibar


def get_pager(pager=None):
    """ Return Pager for use as chat window. """
    from x84.bbs import getterminal, Pager
    term = getterminal()
    height = (term.height - 4)
    width = int(term.width * .9)
    yloc = term.height - height - 1
    xloc = int(term.width / 2) - (width / 2)
    pager = Pager(height, width, yloc, xloc)
    if pager is not None:
        content = pager.content
        # little hack to keep empty lines from re-importing
        for row in range(len(content)):
            ucs = content[row]
            if ucs.startswith(u'\x1b(B'):
                ucs = ucs[len(u'\x1b(B'):]
            if ucs.endswith(u'\x1b[m'):
                ucs = ucs[len(u'\x1b[m'):]
            content[row] = ucs
        pager.update('\r\n'.join(content))
    pager.enable_scrolling = True
    pager.colors['border'] = term.cyan
    pager.glyphs['right-vert'] = u'|'
    pager.glyphs['left-vert'] = u'|'
    pager.glyphs['bot-horiz'] = u''
    return pager


def main(channel=None, caller=None):
    """ Main procedure. """
    # pylint: disable=R0914,R0912,W0603
    #         Too many local variables
    #         Too many branches
    #         Using the global statement
    from x84.bbs import getsession, getterminal, getch, echo
    session, term = getsession(), getterminal()
    global CHANNEL, NICKS
    CHANNEL = '#partyline' if channel is None else channel
    NICKS = dict()

    # sysop repy_to is -1 to force user, otherwise prompt
    if channel == session.sid and caller not in (-1, None):
        echo(u''.join((
            term.normal, u'\a',
            u'\r\n', term.clear_eol,
            u'\r\n', term.clear_eol,
            term.bold_green(u' ** '),
            caller,
            u' would like to chat, accept? ',
            term.bold(u'['),
            term.bold_green_underline(u'yn'),
            term.bold(u']'),
        )))
        while True:
            inp = getch()
            if inp in (u'y', u'Y'):
                break
            elif inp in (u'n', u'N'):
                return False

    def refresh(pager, ipb, init=False):
        """ Returns terminal sequence suitable for refreshing screen. """
        session.activity = 'Chatting in %s' % (
            CHANNEL if not CHANNEL.startswith('#')
            and not 'sysop' in session.user.groups
            else u'PRiVAtE ChANNEl',) if CHANNEL is not None else (
                u'WAitiNG fOR ChAt')
        pager.move_end()
        return u''.join((
            u''.join((u'\r\n', term.clear_eol,
                      u'\r\n', term.clear_eol,
                      term.bold_cyan(u'//'),
                      u' CitZENS bANd'.center(term.width).rstrip(),
                      term.clear_eol,
                       (u'\r\n' + term.clear_eol) * (pager.height + 2),
                      pager.border())) if init else u'',
            pager.title(u''.join((
                term.bold_cyan(u']- '),
                CHANNEL if CHANNEL is not None else u'',
                term.bold_cyan(u' -['),))),
            pager.refresh(),
            ipb.refresh(),))

    def process_cmd(pager, msg):
        """ Process command recieved and display result in chat window. """
        cmd, args = msg.split()[0], msg.split()[1:]
        # pylint: disable=W0603
        #         Using the global statement
        global CHANNEL, NICKS, EXIT
        if cmd.lower() == '/help':
            pager.append(show_help())
            return True
        elif cmd.lower() == '/join' and len(args) == 1:
            part_chan('lEAViNG fOR ANOthER ChANNEl')
            CHANNEL = args[0]
            NICKS = dict()
            join_chan()
            return True
        elif cmd.lower() in ('/act', '/me',):
            act(u' '.join(args))
        elif cmd.lower() == '/say':
            say(u' '.join(args))
        elif cmd.lower() == '/part':
            part_chan(u' '.join(args))
            CHANNEL = None
            NICKS = dict()
            return True
        elif cmd.lower() == '/quit':
            part_chan('quit')
            EXIT = True
        elif cmd.lower() == '/users':
            pager.append(show_nicks(NICKS.keys()))
            return True
        elif cmd.lower() == '/whois' and len(args) == 1:
            whois(args[0])
        return False

    def broadcast_cc(payload):
        """ Broadcast chat even, carbon copy ourselves. """
        session.send_event('global', ('chat', payload))
        session.buffer_event('global', ('chat', payload))

    def join_chan():
        """ Bradcast chat even for /join. """
        payload = (session.sid, CHANNEL, (session.user.handle, 'join', None))
        broadcast_cc(payload)

    def say(mesg):
        """ Signal chat event for /say. """
        payload = (session.sid, CHANNEL, (session.user.handle, 'say', mesg))
        broadcast_cc(payload)

    def act(mesg):
        """ Signal chat event for /act. """
        payload = (session.sid, CHANNEL, (session.user.handle, 'act', mesg))
        broadcast_cc(payload)

    def part_chan(reason):
        """ Signal chat event for /part. """
        payload = (session.sid, CHANNEL, (session.user.handle, 'part', reason))
        broadcast_cc(payload)

    def whois(handle):
        """ Perform /whois request for ``handle``. """
        if not handle in NICKS:
            return
        session.send_event('route', (NICKS[handle], 'info-req', session.sid,))

    def whois_response(attrs):
        """ Display /whois response for given ``attrs``. """
        return show_whois(attrs)

    pager = get_pager(None)  # output window
    readline = get_inputbar(pager)  # input bar
    echo(refresh(pager, readline, init=True))
    echo(pager.append("tYPE '/quit' tO EXit."))
    dirty = time.time()
    join_chan()
    while not EXIT:
        inp = getch(POLL_KEY)

        # poll for and process screen resize
        if session.poll_event('refresh') or (
                inp in (term.KEY_REFRESH, unichr(12))):
            pager = get_pager(pager)
            saved_inp = readline.content
            readline = get_inputbar(pager)
            readline.content = saved_inp
            echo(refresh(pager, readline, init=True))
            dirty = None

        # poll for and process chat events,
        mesg = session.poll_event('global')
        if mesg is not None:
            otxt = process(mesg[1])
            if 0 != len(otxt):
                echo(pager.append(otxt))
                dirty = None if dirty is None else time.time()

        # poll for whois response
        data = session.poll_event('info-ack')
        if data is not None:
            # session id, attributes = data
            echo(pager.append(whois_response(data[1])))
            dirty = None if dirty is None else time.time()

        # process keystroke as input, or, failing that,
        # as a command key to the pager. refresh portions of
        # input bar or act on cariage return, accordingly.
        elif inp is not None:
            otxt = readline.process_keystroke(inp)
            if readline.carriage_returned:
                if readline.content.startswith('/'):
                    if process_cmd(pager, readline.content):
                        pager = get_pager(pager)
                        echo(refresh(pager, readline, init=True))
                elif (0 != len(readline.content.strip())
                        and CHANNEL is not None):
                    say(readline.content)
                readline = get_inputbar(pager)
                echo(readline.refresh())
            elif 0 == len(otxt):
                if type(inp) is int:
                    echo(pager.process_keystroke(inp))
            else:
                echo(u''.join((
                    readline.fixate(-1),
                    readline.colors.get('highlight', u''),
                    otxt, term.normal)))

        # update pager contents. Its a lot for 9600bps ..
        if dirty is not None and time.time() - dirty > POLL_OUT:
            echo(refresh(pager, readline))
            dirty = None
    echo(u''.join((term.move(term.height, 0), term.normal)))
    return True



"""
 New user account script for x/84, https://github.com/jquast/x84
"""


def warning(msg, cpsec=10.0, min_sec=3.0, split_loc=3):
    """
    Display a 2-tone warning to user with a dynamic pause
    """
    from x84.bbs import getterminal, echo, getch
    term = getterminal()
    echo(u''.join((term.clear_eol, term.normal, u'\r\n\r\n',
                   term.bold_red, msg[:-
                                      split_loc], term.normal, msg[
                                          -split_loc:],
                   term.bold_black, u'!')))
    inp = getch(max(min_sec, float(len(msg)) / cpsec))
    return inp


def set_handle(user):
    """
    Prompt for a user.handle, minumum length.
    """
    # pylint: disable=R0914
    #        Too many local variables
    from x84.bbs import getterminal, echo, ini, LineEditor, find_user
    import os  # os.path.sep not allowed in nicks
    import re
    term = getterminal()
    prompt_handle = u'username: '
    msg_empty = u'ENtER AN AliAS'
    msg_exists = u'USER EXiStS.'
    msg_tooshort = u'TOO ShORt, MUSt bE At lEASt %s.'
    msg_invalid = u'IllEGAl USERNAME'
    width = ini.CFG.getint('nua', 'max_user')
    min_user = ini.CFG.getint('nua', 'min_user')
    # pylint: disable=E1103
    #         Instance of '_Chainmap' has no 'split' member
    #         (but some types could not be inferred)
    invalid_nicks = ini.CFG.get('nua', 'invalid_handles').split()
    handle_validation = re.compile(ini.CFG.get('nua', 'handle_validation'))
    while True:
        echo(u'\r\n\r\n' + term.clear_eol + term.normal + prompt_handle)
        user.handle = LineEditor(width, user.handle).read()
        if user.handle == u'' or user.handle is None:
            warning(msg_empty)
        elif find_user(user.handle):
            warning(msg_exists)
        elif len(user.handle) < min_user:
            warning(msg_tooshort % min_user)
        elif ((user.handle.lower() in invalid_nicks)
                or os.path.sep in user.handle):
            warning(msg_invalid)
        elif not handle_validation.match(user.handle):
            warning(msg_invalid)
        else:
            return


def set_location(user):
    """
    Prompt for and set user.location, may be empty
    """
    from x84.bbs import getterminal, echo, ini, LineEditor
    term = getterminal()
    prompt_location = u'origin (optional): '
    width = ini.CFG.getint('nua', 'max_location')
    echo(u'\r\n\r\n' + term.clear_eol + term.normal + prompt_location)
    user.location = LineEditor(width, user.location).read()
    if user.location is None:
        user.location = u''


def set_email(user):
    """
    Prompt for and set user.email, may be empty
    """
    from x84.bbs import getterminal, echo, ini, LineEditor
    term = getterminal()
    prompt_email = u'e-mail (optional): '
    width = ini.CFG.getint('nua', 'max_email')
    echo(u'\r\n\r\n' + term.clear_eol + term.normal + prompt_email)
    user.email = LineEditor(width, user.email).read()
    if user.email is None:
        user.email = u''


def set_password(user):
    """
    Prompt for user.password, minimum length.
    """
    # pylint: disable=R0914
    #        Too many local variables
    from x84.bbs import getterminal, echo, ini, LineEditor
    term = getterminal()
    hidden_ch = u'x'
    prompt_password = u'password: '
    prompt_verify = u'   again: '
    msg_empty = u'ENtER A PASSWORd!'
    msg_tooshort = u'TOO ShORt, MUSt bE At lEASt %s.'
    msg_unmatched = u'VERifY MUSt MAtCH!'
    width = ini.CFG.getint('nua', 'max_pass')
    min_pass = ini.CFG.getint('nua', 'min_pass')
    while True:
        echo(u'\r\n\r\n' + term.clear_eol + term.normal + prompt_password)
        led = LineEditor(width)
        led.hidden = hidden_ch
        password = led.read()
        if password == u'' or password is None:
            warning(msg_empty)
        elif len(password) < min_pass:
            warning(msg_tooshort % min_pass)
        else:
            echo(u'\r\n\r\n' + term.clear_eol + term.normal + prompt_verify)
            led = LineEditor(width)
            led.hidden = hidden_ch
            verify = led.read()
            if password != verify:
                warning(msg_unmatched)
                continue
            user.password = password
            return


def prompt_ok():
    """
    Prompt user to continue, True if they select yes.
    """
    from x84.bbs import getsession, getterminal, echo, getch, Selector
    session, term = getsession(), getterminal()
    prompt_confirm = u'EVERYthiNG lOOk Ok ?'
    prompt_continue = u'YES (CONtiNUE)'
    prompt_chg = u'NO! (ChANGE)'

    def prompt_ok_dumb():
        """ Dummy terminal prompt for confirm/cancel. """
        echo('\r\n\r\n%s\r\n' % (prompt_confirm,))
        echo('1 - %s\r\n' % (prompt_continue,))
        echo('2 - %s\r\n\r\n' % (prompt_chg,))
        echo('select (1, 2) --> ')
        while True:
            inp = getch()
            if inp == u'1':
                return True
            elif inp == u'2':
                return False
    if session.env.get('TERM') == 'unknown':
        return prompt_ok_dumb()
    sel = Selector(yloc=term.height - 1, xloc=5,
                   width=term.width - 10,
                   left=prompt_continue, right=prompt_chg)
    echo(term.normal)
    echo(term.move(term.height - 2, 0) + term.clear_eol)
    echo(prompt_confirm.center(term.width - 1) + '\r\n')
    echo(term.clear_eol + sel.refresh())
    while True:
        echo(sel.process_keystroke(getch()))
        if sel.selected:
            return True if sel.selection == prompt_continue else False


def main(handle=u''):
    """ Main procedure. """
    # pylint: disable=R0914
    #        Too many local variables
    from x84.bbs import getsession, getterminal, echo, ini, User, goto
    from x84.bbs import showcp437
    session, term = getsession(), getterminal()
    import os
    session.activity = u'Applying for an account'
    artfile = os.path.join(os.path.dirname(__file__), 'art', 'nua.asc')
    msg_header = u'NEW USER APPliCAtiON'
    # pylint: disable=E1103
    #         Instance of '_Chainmap' has no 'split' member
    #         (but some types could not be inferred)
    newcmds = ini.CFG.get('matrix', 'newcmds').split()
    topscript = ini.CFG.get('matrix', 'topscript')

    # display art and msg_header as banner
    echo(u'\r\n\r\n')
    for line in showcp437(artfile):
        echo(line)
    echo(u'\r\n\r\n' + term.reverse + msg_header.center(term.width))

    # create new user record for manipulation
    user = User(handle if handle.lower() not in newcmds else u'')
    while True:
        set_handle(user)
        set_location(user)
        set_email(user)
        set_password(user)
        if prompt_ok():
            user.save()
            goto(topscript, user.handle)



"""
Logging handler for x/84 BBS, http://github.com/jquast/x84
"""
import logging
import copy
import sys
from x84.blessings import Terminal


class ColoredConsoleHandler(logging.StreamHandler):
    """
    A stream handler that colors the levelname, thats all.
    """

    def __init__(self):
        """ Constructor class, initializes blessings Terminal. """
        self.term = Terminal(stream=sys.stderr)
        logging.StreamHandler.__init__(self)

    def color_levelname(self, record):
        """ Modify levelname field to include terminal color sequences.  """
        record.levelname = (self.term.bold_red if record.levelno >= 50 else
                            self.term.bold_red if record.levelno >= 40 else
                            self.term.bold_yellow if record.levelno >= 30 else
                            self.term.bold_white if record.levelno >= 20 else
                            self.term.blue)('%-5s' % (
                                record.levelname.title()
                                if record.levelname.lower() != 'warning'
                                else 'warn',))
        return record

    def transform(self, src_record):
        """ Return a modified log record """
        return self.color_levelname(src_record)

    def emit(self, src_record):
        """ Emit record to console """
        # transform
        dst_record = self.transform(copy.copy(src_record))

        # emit to console
        logging.StreamHandler.emit(self, dst_record)



""" oneliners for x/84, http://github.com/jquast/x84

  To use the (optional) http://bbs-scene.org API,
  configure a section in your .ini file:

    [bbs-scene]
    user = my@email-addr.ess
    pass = my-plaintext-password
"""
import threading

# bbs-scene.org API is 50-character limit max
MAX_INPUT = 50
# scrollback / database limit
BUF_HISTORY = 500
# XML/API requests limit
XML_HISTORY = 84
# in dumb terminal mode or expert mode, wait up to
# this long for bbs-scene.org updates ('x' cancels)
WAIT_FETCH = 5


class FetchUpdates(threading.Thread):
    """ Fetch bbs-scene.org oneliners as a background thread. """
    url = 'http://bbs-scene.org/api/onelinerz?limit=%d' % (XML_HISTORY,)
    content = list()

    def run(self):
        import logging
        import time
        import requests
        import xml.etree.ElementTree
        from x84.bbs import ini
        logger = logging.getLogger()
        usernm = ini.CFG.get('bbs-scene', 'user')
        passwd = ini.CFG.get('bbs-scene', 'pass')
        logger.debug('fetching %r ..', self.url)
        stime = time.time()
        req = requests.get(self.url, auth=(usernm, passwd))
        if 200 != req.status_code:
            logger.error(req.content)
            logger.error('bbs-scene.org returned %s', req.status_code)
            return
        else:
            logger.info('bbs-scene.org returned %d in %2.2fs',
                        req.status_code, time.time() - stime)

        # can throw exceptions when xml is invalid, as a thread, nobody needs
        # to catch it. theres some things that should be CDATA wrapped ..
        buf = ''.join((byte for byte in req.content
                       if ord(byte) >= 0x20
                       or ord(byte) in (0x09, 0x0a, 0x0d, 0x7f)))
        xml_nodes = xml.etree.ElementTree.XML(buf).findall('node')
        for node in xml_nodes:
            self.content.append(
                (node.find('id').text, dict(
                    ((key, (node.find(key).text
                            if (node.find(key) is not None and
                                node.find(key).text is not None)
                            else u'').strip()) for key in
                     ('oneliner', 'alias', 'bbsname', 'timestamp',))),))


def wait_for(thread):
    """
    for dummy or small terminals, wait until a request thread has
    gotten content before continuing, up to t WAIT_FETCH seconds.
    """
    from x84.bbs import echo, getch
    if thread.is_alive():
        echo(u"\r\n\r\nfetching bbs-scene.org oneliners .. "
             "(%s)s\b\b%s" % (' ' * 2, '\b' * 2,))
        for num in range(WAIT_FETCH):
            echo(u'%2d%s' % (WAIT_FETCH - num - 1, '\b' * 2,))
            if not thread.is_alive():
                return
            thread.join(1)  # block 1 second on thread
            if getch(0) == u'q':
                return     # allow cancel using 'q'


def chk_thread(thread):
    """
    check if bbs-scene.org thread finished, if so, farm
    its data and send updates via event 'oneliner_update' if there
    are any.
    """
    from x84.bbs import getsession, DBProxy
    import logging
    logger = logging.getLogger()
    session = getsession()
    if thread is not None and not thread.is_alive():
        udb = DBProxy('oneliner')
        udbkeys = udb.keys()
        nlc = 0
        for key, value in thread.content:
            if key not in udbkeys:
                udb[key] = value
                nlc += 1
        if nlc:
            logger.debug('%d new entries', nlc)
            session.buffer_event('oneliner_update', True)
        else:
            logger.debug('no new bbs-scene.org entries')
        return True


def add_oneline(msg):
    """
    Add a oneliner to the local database.
    """
    import time
    from x84.bbs import getsession, DBProxy, ini
    session = getsession()
    udb = DBProxy('oneliner')
    udb.acquire()
    udb[max([int(key) for key in udb.keys()] or [0]) + 1] = {
        'oneliner': msg,
        'alias': getsession().handle,
        'bbsname': ini.CFG.get('system', 'bbsname'),
        'timestamp': time.strftime('%Y-%m-%d %H:%M:%S'),
    }
    udb.release()
    session.buffer_event('oneliner_update', True)
    session.send_event('global', ('oneliner_update', True))


def get_oltxt():
    """ Return unicode terminal string of oneliners. """
    import time
    from x84.bbs import getterminal, DBProxy, timeago, Ansi
    term = getterminal()
    colors = (term.bold_white, term.bold_green, term.bold_blue)
    hist = [(int(k), v) for (k, v) in DBProxy('oneliner').items()]
    hist.sort()
    output = list()
    for idx, onel in hist[BUF_HISTORY * -1:]:
        color = colors[int(idx) % len(colors)]
        atime = timeago(time.time() - time.mktime(
            time.strptime(onel['timestamp'], '%Y-%m-%d %H:%M:%S'))).strip()
        output.append(u''.join((
            term.bold_white('('),
            color(atime), term.bold_black(u' ago'),
            term.bold_black(u' '),
            color(onel['alias']),
            term.bold_black(u'/'), onel['bbsname'],
            term.bold_white(u')'), color(u': '),
            Ansi(onel['oneliner']).decode_pipe(),
        )))
    return output[(BUF_HISTORY * -1):]


def get_selector(selection=u'No'):
    """ Return yes/no selector """
    from x84.bbs import getterminal, Selector
    term = getterminal()
    selector = Selector(
        yloc=term.height - 1, xloc=(term.width / 2) - 25,
        width=50, left=u'Yes', right=u'No')
    selector.keyset['left'].extend((u'y', u'Y'))
    selector.keyset['right'].extend((u'y', u'Y'))
    selector.selection = selection
    selector.colors['selected'] = term.green_reverse
    return selector


def get_pager():
    """ Returns pager window for oneliners content. """
    from x84.bbs import getterminal, Pager
    term = getterminal()
    xloc = max(3, (term.width / 2) - 50)
    width = min(term.width - 6, 100)
    pager = Pager(yloc=9, xloc=xloc, height=term.height - 12, width=width)
    pager.colors['border'] = term.blue
    return pager


def banner():
    """ Return banner """
    from x84.bbs import getterminal, Ansi, from_cp437
    import os
    term = getterminal()
    output = u''
    output += u'\r\n\r\n'
    if term.width >= 78:
        output += term.home + term.normal + term.clear
        # xzip's ansi is line-clean, center-align with terminal width,
        artfile = os.path.join(os.path.dirname(__file__), 'ol.ans')
        art = open(artfile).readlines()
        max_ans = max([len(Ansi(from_cp437(line.rstrip()))) for line in art])
        for line in art:
            padded = Ansi(from_cp437(line.rstrip())).center(max_ans)
            output += term.normal + term.blue  # minor fix for this art ;/
            output += Ansi(padded).center(term.width).rstrip() + '\r\n'
    return output + term.normal


def redraw(pager, selector):
    """ Redraw pager and selector """
    from x84.bbs import getsession, getterminal, echo
    session, term = getsession(), getterminal()
    session.flush_event('oneliner_update')
    pager.colors['border'] = term.white
    pager.glyphs['left-vert'] = u' '
    pager.glyphs['right-vert'] = u' '
    prompt_ole = u'SAY somethiNG ?!'
    pager.update(u'\n\n\nFetching ...')
    echo(u''.join((
        pager.refresh(),
        pager.border(),
        term.move(selector.yloc - 2, selector.xloc),
        term.bold_green(prompt_ole.center(selector.width).rstrip()),
        term.clear_eol,
        selector.refresh(),)))
    pager.update(u'\n'.join(get_oltxt()))
    pager.move_end()
    echo(pager.refresh())


def dummy_pager():
    """ Display oneliners for dummy terminals and prompt to saysomething """
    from x84.bbs import getterminal, Ansi, echo, getch
    term = getterminal()
    tail = term.height - 4
    indent = 3
    prompt_ole = u'\r\n\r\nSAY somethiNG ?! [yn]'
    buf = list()
    for record in get_oltxt():
        buf.extend(Ansi(record.rstrip()).wrap(
            term.width - indent).split('\r\n'))
    echo((u'\r\n' + term.normal).join(buf[tail * -1:]))
    echo(prompt_ole)
    while True:
        inp = getch()
        if inp in (u'n', u'N', u'q', term.KEY_EXIT):
            break
        if inp in (u'y', u'Y'):
            return saysomething(dumb=True)
    return


def saysomething(dumb=True):
    """
    Prompt user to post oneliner, also prompt user to post
    to bbs-scene.org if configured, returning background Thread.
    """
    import time
    from x84.bbs import getsession, getterminal, echo, LineEditor, ini
    session, term = getsession(), getterminal()
    prompt_say = u'SAY WhAt ?! '
    # heard_msg = u'YOUR MESSAGE hAS bEEN VOiCEd.'

    yloc = term.height - 3
    xloc = max(0, ((term.width / 2) - (MAX_INPUT / 2)))
    if dumb:
        echo(u'\r\n\r\n' + term.bold_blue(prompt_say))
    else:
        echo(term.move(yloc, xloc) or u'\r\n\r\n')
        echo(term.bold_blue(prompt_say))
    ole = LineEditor(MAX_INPUT)
    ole.highlight = term.green_reverse
    oneliner = ole.read()
    if oneliner is None or 0 == len(oneliner.strip()):
        if not dumb:
            # clear input line,
            echo(term.normal + term.move(yloc, 0) + term.clear_eol)
        return None

    session.user['lastliner'] = time.time()
    # post local-onlyw hen bbs-scene.org is not configured
    if not ini.CFG.has_section('bbs-scene'):
        add_oneline(oneliner.strip())
        return None
    return post_bbs_scene(oneliner, dumb)


def post_bbs_scene(oneliner, dumb=True):
    """
    Prompt for posting to bbs-scene.org oneliners API,
    returning thread if posting occured.
    """
    # pylint: disable=R0914
    #        Too many local variables
    import logging
    import xml.etree.ElementTree
    import requests
    from x84.bbs import echo, getch, getterminal, getsession, ini
    logger = logging.getLogger()
    session, term = getsession(), getterminal()
    prompt_api = u'MAkE AN ASS Of YOURSElf ON bbS-SCENE.ORG?!'
    heard_api = u'YOUR MESSAGE hAS bEEN brOAdCAStEd.'
    yloc = term.height - 3
    if dumb:
        # post to bbs-scene.org ?
        echo('\r\n\r\n' + term.bold_blue(prompt_api) + u' [yn]')
        inp = getch(1)
        while inp not in (u'y', u'Y', u'n', u'N'):
            inp = getch()
        if inp in (u'n', u'N'):
            #  no? then just post locally
            add_oneline(oneliner.strip())
            return None
    else:
        # fancy prompt, 'post to bbs-scene.org?'
        sel = get_selector()
        sel.colors['selected'] = term.red_reverse
        echo(term.move(sel.yloc - 1, sel.xloc) or ('\r\n\r\n'))
        echo(term.blue_reverse(prompt_api.center(sel.width)))
        echo(sel.refresh())
        while not sel.selected and not sel.quit:
            echo(sel.process_keystroke(getch()))
        session.buffer_event('refresh', 'dirty')
        if sel.quit or sel.selection == sel.right:
            echo(term.normal + term.move(yloc, 0) + term.clear_eol)
            echo(term.move(sel.yloc, 0) + term.clear_eol)
            return None

    # This is an AJAX effect.
    # Dispatch a thread to fetch updates, whose callback
    # will cause the database and pager to update.
    url = 'http://bbs-scene.org/api/onelinerz.xml'
    usernm = ini.CFG.get('bbs-scene', 'user')
    passwd = ini.CFG.get('bbs-scene', 'pass')
    data = {
            'bbsname': ini.CFG.get('system', 'bbsname'),
            'alias': session.user.handle,
            'oneliner': oneliner.strip(),
            }
    # post to bbs-scene.rog
    req = requests.post(url, auth=(usernm, passwd), data=data)
    if (req.status_code != 200 or
            (xml.etree.ElementTree.XML(req.content)
                .find('success').text != 'true')):
        echo(u'\r\n\r\n%srequest failed,\r\n' % (term.clear_eol,))
        echo(u'%r' % (req.content,))
        echo(u'\r\n\r\n%s(code: %s).\r\n' % (
            term.clear_eol, req.status_code,))
        echo(u'\r\n%sPress any key ..' % (term.clear_eol,))
        logger.warn('bbs-scene.org api request failed')
        getch()
        return None
    logger.info('bbs-scene.org api (%d): %r/%r', req.status_code,
                session.user.handle, oneliner.strip())
    thread = FetchUpdates()
    thread.start()
    if not dumb:
        # clear line w/input bar,
        echo(term.normal + term.move(yloc, 0) + term.clear_eol)
        # clear line w/selector
        echo(term.move(sel.yloc, 0) + term.clear_eol)
    else:
        echo('\r\n\r\n' + heard_api)
        getch(2)
    return thread


def main():
    """ Main procedure. """
    # pylint: disable=R0912
    #        Too many branches
    from x84.bbs import getsession, getterminal, ini, echo, getch
    session, term = getsession(), getterminal()
    pager, selector = get_pager(), get_selector()

    thread = None
    if ini.CFG.has_section('bbs-scene'):
        thread = FetchUpdates()
        thread.start()
        session.activity = u'one-liners [bbs-scene.org]'
    else:
        session.activity = u'one-liners'

    # flag a pager update,
    dirty = True
    # force screen clear on first loop,
    session.buffer_event('refresh', ('init',))
    while True:
        # 1. calculate and redraw screen,
        # or enter dumb pager mode (no scrolling)
        if session.poll_event('refresh'):
            pager, selector = get_pager(), get_selector(selector.selection)
            echo(banner())
            dirty = True
        if chk_thread(thread):
            thread = None
        while session.read_event('oneliner_update', 0.15):
            dirty = True
        if dirty and (session.env.get('TERM') != 'unknown'
                      and not session.user.get('expert', False)
                      and term.width >= 78 and term.height >= 20):
            # smart terminal
            redraw(pager, selector)
            dirty = False
        elif dirty:
            # dumb terminal
            if thread is not None:
                wait_for(thread)
            if chk_thread(thread):
                thread = None
            echo(u'\r\n\r\n')
            return dummy_pager()

        # 2. detect and process keyboard input,
        inp = getch(1)
        if inp is not None:
            # input is multiplexed to both interfaces
            echo(pager.process_keystroke(inp))
            echo(selector.process_keystroke(inp))

            # selected 'yes' & return, 'say something'
            if (selector.selected and selector.selection == selector.left):
                # re-assign thread so that it is checked for updates
                thread = saysomething(dumb=False)
                # undo 'selected' state of yes/no bar,
                selector.selected = False

            # quit 'q', or selected 'no' & return
            elif (selector.selected and selector.selection == selector.right
                    or pager.quit):
                return



"""
 User profile editor script for x/84, http://github.com/jquast/x84
"""


ABOUT_DOT_PLAN = (u'The .plan file is a throwback to early Unix '
                  + u'"blogosphere", this is a simple file that is '
                  + u'GLOBALLY shared with all other users. You can '
                  + u'put anything you want here: something about '
                  + u'yourself and your interests, your websites, '
                  + u'greetz, etc.')

ABOUT_TERM = (u'This only sets TERM for new processes, such as doors. '
              + u'Your bbs session itself discovers TERM only once during '
              + u'telnet negotiation on-connect.')

EXIT = False


def process_keystroke(inp, user):
    """ Process keystroke, ``inp``, for target ``user``. """
    # pylint: disable=R0914,R0912,R0915,R0911,W0603
    #         Too many local variables
    #         Too many branches
    #         Too many statements
    #         Too many return statements
    #         Using the global statement
    # ^ lol, this is one of those things that should be
    #   refactored into smaller subroutines =)
    from x84.bbs import getsession, getterminal, echo, getch, gosub
    from x84.bbs import LineEditor, Ansi
    from x84.default.nua import set_email, set_location
    from x84.bbs.ini import CFG
    def_timeout = CFG.getint('system', 'timeout')
    global EXIT
    session, term = getsession(), getterminal()
    is_self = bool(user.handle == session.user.handle)
    invalid = u'\r\niNVAlid.'
    assert is_self or 'sysop' in session.user.groups

    if is_self and inp in (u'c', u'C'):
        gosub('charset')

    elif is_self and inp in (u't', u'T'):
        echo(term.move(term.height - 1, 0))
        echo(ABOUT_TERM + u'\r\n')
        echo(u'\r\ntERMiNAl tYPE: ')
        term = LineEditor(30, session.env.get('TERM')).read()
        echo(u"\r\n\r\nset TERM to '%s'? [yn]" % (term,))
        while True:
            inp2 = getch()
            if inp2 in (u'y', u'Y'):
                session.env['TERM'] = term
                break
            elif inp2 in (u'n', u'N'):
                break

    elif is_self and inp in (u'w', u'W'):
        echo(u'\r\ntERMiNAl Width: ')
        width = LineEditor(3, str(term.width)).read()
        try:
            width = int(width)
        except ValueError:
            echo(invalid)
            return True
        if width < 0 or width > 999:
            echo(invalid)
            return True
        echo(u"\r\n\r\nset COLUMNS=%d? [yn]" % (width,))
        while True:
            inp2 = getch()
            if inp2 in (u'y', u'Y'):
                term.columns = width
                break
            elif inp2 in (u'n', u'N'):
                break

    elif is_self and inp in (u'h', u'H'):
        echo(u'\r\ntERMiNAl hEiGht: ')
        height = LineEditor(3, str(term.height)).read()
        try:
            height = int(height)
        except ValueError:
            echo(invalid)
            return True
        if height < 0 or height > 999:
            echo(invalid)
            return True
        echo(u"\r\n\r\nset LINES=%d? [yn]" % (height,))
        while True:
            inp2 = getch()
            if inp2 in (u'y', u'Y'):
                term.rows = height
                break
            elif inp2 in (u'n', u'N'):
                break

    elif 'sysop' in session.user.groups and inp in (u'd', u'D',):
        echo(u"\r\n\r\ndElEtE %s ? [yn]" % (user.handle,))
        while True:
            inp2 = getch()
            if inp2 in (u'y', u'Y'):
                user.delete()
                break
            elif inp2 in (u'n', u'N'):
                break
        EXIT = True

    elif 'sysop' in session.user.groups and inp in (u's', u'S',):
        sysop = not 'sysop' in user.groups
        echo(u"\r\n\r\n%s SYSOP ACCESS? [yn]" % (
            'ENAblE' if sysop else 'diSAblE',))
        while True:
            inp2 = getch()
            if inp2 in (u'y', u'Y'):
                if sysop:
                    user.groups.add('sysop')
                else:
                    user.groups.remove('sysop')
                user.save()
                break
            elif inp2 in (u'n', u'N'):
                break
    elif inp in (u'p', u'P'):
        from x84.default.nua import set_password
        set_password(user)
        echo(u"\r\n\r\nSEt PASSWORd ? [yn]")
        while True:
            inp2 = getch()
            if inp2 in (u'y', u'Y'):
                user.save()
                break
            elif inp2 in (u'n', u'N'):
                break
    elif inp in (u'.',):
        echo(term.move(0, 0) + term.normal + term.clear)
        echo(term.move(int(term.height * .8), 0))
        for line in Ansi(ABOUT_DOT_PLAN).wrap(
                term.width / 3).splitlines():
            echo(line.center(term.width).rstrip() + u'\r\n')
        echo(u'\r\n\r\nPRESS ANY kEY ...')
        getch()
        if is_self:
            gosub('editor', '.plan')
        else:
            tmpkey = '%s-%s' % (user.handle, user.plan)
            draft = user.get('.plan', u'')
            session.user[tmpkey] = draft
            gosub('editor', tmpkey)
            if session.user.get(tmpkey, u'') != draft:
                echo(u"\r\n\r\nSEt .PlAN ? [yn]")
                while True:
                    inp2 = getch()
                    if inp2 in (u'y', u'Y'):
                        user['.plan'] = session.user[tmpkey]
                        break
                    elif inp2 in (u'n', u'N'):
                        break
    elif inp in (u'l', u'L'):
        echo(term.move(term.height - 1, 0))
        set_location(user)
        echo(u"\r\n\r\nSEt lOCAtiON tO '%s'? [yn]" % (user.location,))
        while True:
            inp2 = getch()
            if inp2 in (u'y', u'Y'):
                user.save()
                break
            elif inp2 in (u'n', u'N'):
                break
    elif inp in (u'e', u'E'):
        echo(term.move(term.height - 1, 0))
        set_email(user)
        echo(u"\r\n\r\nSEt EMAil tO '%s'? [yn]" % (user.email,))
        while True:
            inp2 = getch()
            if inp2 in (u'y', u'Y'):
                user.save()
                break
            elif inp2 in (u'n', u'N'):
                break
    elif inp in (u'i', u'I'):
        echo(u'\r\ntiMEOUt (0=NONE): ')
        timeout = LineEditor(6, str(user.get('timeout', def_timeout))).read()
        try:
            timeout = int(timeout)
        except ValueError:
            echo(invalid)
            return True
        if timeout < 0:
            echo(invalid)
            return True
        echo(u"\r\n\r\nSet tiMEOUt=%s? [yn]" % (
            timeout if timeout else 'None',))
        while True:
            inp2 = getch()
            if inp2 in (u'y', u'Y'):
                user['timeout'] = timeout
                session.send_event('set-timeout', timeout)
                break
            elif inp2 in (u'n', u'N'):
                break

    elif inp in (u'm', u'M'):
        mesg = False if user.get('mesg', True) else True
        echo(u"\r\n\r\n%s iNStANt MESSAGiNG? [yn]" % (
            'ENAblE' if mesg else 'DiSAblE',))
        while True:
            inp2 = getch()
            if inp2 in (u'y', u'Y'):
                user['mesg'] = mesg
                break
            elif inp2 in (u'n', u'N'):
                break
    elif inp in (u'x', u'X'):
        expert = not user.get('expert', False)
        echo(u"\r\n\r\n%s EXPERt MOdE? [yn]" % (
            'ENAblE' if expert else 'DiSAblE',))
        while True:
            inp2 = getch()
            if inp2 in (u'y', u'Y'):
                user['expert'] = expert
                break
            elif inp2 in (u'n', u'N'):
                break
    elif inp in (u'q', u'Q',):
        EXIT = True
    else:
        return False
    return True


def dummy_pager(user):
    """ A dummy selector for profile attributes """
    from x84.bbs import getsession, getterminal, echo, Ansi, getch
    session, term = getsession(), getterminal()
    plan = user.get('.plan', False)
    from x84.bbs.ini import CFG
    def_timeout = CFG.getint('system', 'timeout')
    menu = ['(c)%-20s - %s' % (u'hARACtER ENCOdiNG',
                               term.bold(session.encoding),),
            '(t)%-20s - %s' % (u'ERMiNAl tYPE',
                               term.bold(session.env.get('TERM', 'unknown')),),
            '(h)%-20s - %s' % (u'ERMiNAl hEiGht',
                               term.bold(str(term.height)),),
            '(w)%-20s - %s' % (u'ERMiNAl WidtH',
                               term.bold(str(term.width)),),
            '(l)%-20s - %s' % (u'OCAtiON',
                               term.bold(user.location),),
            '(p)%-20s - %s' % (u'ASSWORd',
                               term.bold_black(u'******'),),
            '(e)%-20s - %s' % (u'-MAil AddRESS',
                               term.bold(user.email),),
            (term.bold('t') +
                '(i)%-19s - %s' % (u'MEOUt', term.bold(
                    str(user.get('timeout', def_timeout))),)),
            '(s)%-20s - %s' % (u'YSOP ACCESS',
                               term.bold(u'ENAblEd'
                                         if 'sysop' in user.groups
                                         else 'diSAblEd')),
            '(m)%-20s - %s' % (u'ESG',
                               term.bold(u'[%s]' % ('y'
                                   if user.get(
                                       'mesg', True) else 'n',),)),
            '(.)%-20s - %s' % (u'PlAN filE', '%d bytes' % (
                len(plan),) if plan else '(NO PlAN.)'),
            '(x)%-20s - %s' % (u'PERt MOdE',
                               term.bold(u'ENAblEd'
                                         if user.get('expert', False)
                                         else 'diSAblEd')),
            '(q)Uit', ]
    echo(term.normal + u'\r\n\r\n')
    lines = Ansi('\n'.join(menu)).wrap(term.width).splitlines()
    xpos = max(1, int(term.width / 2) - (40 / 2))
    for row, line in enumerate(lines):
        if row and (0 == row % (term.height - 2)):
            echo(term.reverse(u'\r\n-- More --'))
            getch()
        echo(u'\r\n' + ' ' * xpos + line)
    echo(u'\r\n\r\n Enter option [ctle.xq]: ')
    return process_keystroke(getch(), user)


def main(handle=None):
    """ Main procedure. """
    # pylint: disable=W0603
    #         Using the global statement
    from x84.bbs import getsession, getterminal
    from x84.bbs import get_user
    session, term = getsession(), getterminal()
    user = session.user if ('sysop' not in session.user.groups
            ) or (handle is None) else get_user(handle)
    global EXIT
    while not EXIT:
        session.activity = 'User profile editor'
        dummy_pager(user)



"""
x/84 bbs module, https://github.com/jquast/x84
"""


from x84.bbs.userbase import list_users, get_user, find_user, User, Group
from x84.bbs.msgbase import list_msgs, get_msg, list_tags, Msg
from x84.bbs.exception import Disconnected, Goto
from x84.bbs.editor import LineEditor, ScrollingEditor
from x84.bbs.output import echo, timeago, Ansi, ansiwrap
from x84.bbs.ansiwin import AnsiWindow
from x84.bbs.selector import Selector
from x84.bbs.lightbar import Lightbar
from x84.bbs.cp437 import from_cp437
from x84.bbs.dbproxy import DBProxy
from x84.bbs.pager import Pager
from x84.bbs.door import Door, DOSDoor, Dropfile

__all__ = ['list_users', 'get_user', 'find_user', 'User', 'Group',
    'list_msgs', 'get_msg', 'list_tags', 'Msg',
    'LineEditor', 'ScrollingEditor', 'echo', 'timeago', 'Ansi',
    'ansiwrap', 'AnsiWindow', 'Selector', 'Lightbar', 'from_cp437',
    'DBProxy', 'Pager', 'Door', 'DOSDoor', 'goto', 'disconnect',
    'getsession', 'getterminal', 'getch', 'gosub', 'ropen',
    'showcp437', 'Dropfile',]

def goto(*args):
    """
    Change bbs script. Does not return.
    """
    raise Goto(args)


def disconnect(reason=u''):
    """
    Disconnect session. Does not return.
    """
    raise Disconnected(reason,)


def getsession():
    """
    Returns Session of calling process.
    """
    import x84.bbs.session
    return x84.bbs.session.getsession()


def getterminal():
    """
    Returns Terminal of calling process.
    """
    import x84.bbs.session
    return x84.bbs.session.getterminal()


# temporary hacks until blessings updates with term.inkey() upstream ..
def getch(timeout=None):
    """
    Retrieve a keystroke from 'input' queue, blocking forever or, when
    specified, None when timeout has elapsed.

    upstream blessings has better 'keycode' evaluation (none of this
    duck typing, its always unicode, but has .is_sequence bool test,
    and a .value test for keycode comparison). we workaround for legacy
    behavior unless upstream blessings accepts our impl. in some form ..
    """
    keystroke = getterminal().inkey(timeout)
    if keystroke is None:
        return None
    if keystroke.is_sequence:
        return keystroke.code
    return keystroke

def gosub(*arg):
    """
    Call bbs script with optional arguments, Returns value.
    """
    # pylint: disable=W0142
    #        Used * or ** magic
    return getsession().runscript(*(arg[0],) + arg[1:])


def ropen(filename, mode='rb'):
    """
    Open random file using wildcard (glob)
    """
    import glob
    import random
    files = glob.glob(filename)
    return open(random.choice(files), mode) if len(files) else None


def showcp437(filepattern):
    """
    yield unicode sequences for any given ANSI Art (of cp437 encoding). Effort
    is made to strip SAUCE data, translate cp437 to unicode, and trim artwork
    too large to display. If keyboard input is pressed, 'msg_cancel' is
    returned as the last line of art
    """
    import sauce
    session, term = getsession(), getterminal()
    msg_cancel = u''.join((term.normal,
                           term.bold_blue(u'--'),
                           u'CANCEllEd bY iNPUt ',
                           term.bold_blue(u'--'),))
    msg_notfound = u''.join((
        term.bold_red(u'--'),
        u'no files matching %s' % (filepattern,),
        term.bold_red(u'--'),))
    fobj = (ropen(filepattern)
            if '*' in filepattern or '?' in filepattern
            else open(filepattern, 'rb'))
    if fobj is None:
        yield msg_notfound + u'\r\n'
        return
    # allow slow terminals to cancel by pressing a keystroke
    for line in from_cp437(sauce.SAUCE(fobj).__str__()).splitlines():
        if session.poll_event('input'):
            yield u'\r\n' + msg_cancel + u'\r\n'
            return
        yield line + u'\r\n'




"""
Post-login screen for x/84, http://github.com/jquast/x84

When handle is None or u'', an in-memory account 'anonymous' is created
and assigned to the session.
"""

# generated using lolcat ..

BADGE256 = u'\x1b'.join((u'',
                            u'[38;5;49m2', u'[0m', u'[38;5;48m5', u'[0m',
                            u'[38;5;48m6', u'[0m', u'[38;5;48m-', u'[0m',
                            u'[38;5;48mC', u'[0m', u'[38;5;84mO', u'[0m',
                            u'[38;5;83ml', u'[0m', u'[38;5;83mO', u'[0m',
                            u'[38;5;83mr', u'[0m', u'[38;5;83m ', u'[0m',
                            u'[38;5;83mb', u'[0m', u'[38;5;119mA', u'[0m',
                            u'[38;5;118md', u'[0m', u'[38;5;118mG', u'[0m',
                            u'[38;5;118mE', u'[0m', u'[38;5;118m ', u'[0m',
                            u'[38;5;154mA', u'[0m', u'[38;5;154mW', u'[0m',
                            u'[38;5;154mA', u'[0m', u'[38;5;154mR', u'[0m',
                            u'[38;5;154mD', u'[0m', u'[38;5;184mE', u'[0m',
                            u'[38;5;184md', u'[0m', u'[38;5;184m!', u'[0m',))


def display_intro():
    """ Display art, 256-color badge, and '!'encoding help. """
    from x84.bbs import getterminal, showcp437, echo
    import os
    term = getterminal()
    # display random artwork ..
    artfile = os.path.join(os.path.dirname(__file__), 'art',
                           '*.ans' if term.number_of_colors != 0 else '*.asc')
    echo(u'\r\n\r\n')
    for line in showcp437(artfile):
        echo(line)
    if term.number_of_colors == 256 and term.kind.startswith('xterm'):
        echo(u''.join((
            term.normal, '\r\n\r\n',
            BADGE256, u'\r\n',
            u'\r\n')))
    echo(u'! to change encoding\r\n')


def get_ynbar():
    """ Retrieve yes/no bar for quick login. """
    from x84.bbs import getterminal, Selector
    term = getterminal()
    ynbar = Selector(yloc=term.height - 1,
                     xloc=term.width - 31,
                     width=30, left='Yes', right='No')
    ynbar.colors['selected'] = term.green_reverse
    ynbar.keyset['left'].extend((u'y', u'Y',))
    ynbar.keyset['right'].extend((u'n', u'N',))
    return ynbar


def redraw_quicklogin(ynbar):
    """ Redraq yes/no bar for quick login. """
    from x84.bbs import getterminal
    prompt_ql = u' QUiCk lOGiN ?! '
    term = getterminal()
    return u''.join((
        term.move(ynbar.yloc - 1, ynbar.xloc),
        term.normal,
        term.bold_blue(prompt_ql),
        ynbar.refresh(),
    ))


def main(handle=None):
    """ Main procedure. """
    # pylint: disable=R0914,R0912,R0915
    #         Too many local variables
    #         Too many branches
    #         Too many statements
    from x84.bbs import getsession, getterminal, echo, getch
    from x84.bbs import goto, gosub, User, get_user, DBProxy
    import logging
    import time
    session, term = getsession(), getterminal()
    session.activity = 'top'
    logger = logging.getLogger()

    # 0. just a gimmicky example,
    gosub('productive')

    # 1. determine & assign user record,
    if handle in (None, u'', 'anonymous',):
        logger.info('anonymous login by %s.', session.sid)
        session.user = User(u'anonymous')
    else:
        logger.debug('%r logged in.', handle)
        session.user = get_user(handle)
        timeout = session.user.get('timeout', None)
        if timeout is not None:
            echo(u'\r\n\r\nUsing preferred timeout of %ss.\r\n' % (
                timeout,))
            session.send_event('set-timeout', timeout)

    # 2. update call records
    session.user.calls += 1
    session.user.lastcall = time.time()
    if session.user.handle != 'anonymous':
        session.user.save()

    # record into " last caller " record
    key = (session.user.handle)
    lcall = (session.user.lastcall, session.user.calls, session.user.location)
    db = DBProxy('lastcalls')
    db[key] = lcall

    # 3. if no preferred charset run charset.py selector
    if (session.user.get('charset', None) is None
            or session.user.handle == 'anonymous'):
        gosub('charset')
        session.activity = 'top'
    else:
        # load default charset
        session.encoding = session.user.get('charset')
        fun = term.bold_green(' (EXCEllENt!)')
        if session.encoding != 'utf8':
            fun = term.bold_red(u' (bUMMER!)')
        echo(u'\r\n\r\nUsing preferred charset, %s%s.\r\n' % (
            session.encoding, fun))

    # 4. impress with art, prompt for quick login (goto 'main'),
    if session.user.get('expert', False):
        dirty = True
        while True:
            if session.poll_event('refresh'):
                dirty = True
            if dirty:
                session.activity = 'top'
                display_intro()
                echo(u'\r\n QUiCk lOGiN [yn] ?\b\b')
            dirty = False
            inp = getch(1)
            if inp in (u'y', u'Y'):
                goto('main')
            elif inp in (u'n', u'N'):
                break
            elif inp in (u'!',):
                gosub('charset')
                dirty = True
    else:
        ynbar = get_ynbar()
        dirty = True
        while not ynbar.selected:
            if session.poll_event('refresh'):
                dirty = True
            if dirty:
                # redraw yes/no
                session.activity = 'top'
                swp = ynbar.selection
                ynbar = get_ynbar()
                ynbar.selection = swp
                display_intro()
                echo(redraw_quicklogin(ynbar))
            dirty = False
            inp = getch(1)
            if inp in (u'!',):
                gosub('charset')
                dirty = True
            elif inp is not None:
                echo(ynbar.process_keystroke(inp))
                if ynbar.quit:
                    goto('main')
        if ynbar.selection == ynbar.left:
            goto('main')

    # 5. last callers
    gosub('lc')
    session.activity = 'top'

    # 6. check for new public/private msgs,
    gosub('readmsgs', set())
    session.activity = 'top'

    # 7. news
    gosub('news')
    session.activity = 'top'

    # 8. one-liners
    gosub('ol')
    session.activity = 'top'

    # 9. weather
    if session.user.get('location', None):
        gosub('weather')

    goto('main')



"""
Left/Right lightbar choice selector for x/84, https://github.com/jquast/x84
"""
from x84.bbs.ansiwin import AnsiWindow

VI_KEYSET = {
    'refresh': [unichr(12)],
    'toggle': [u' '],
    'left': [u'h'],
    'right': [u'l'],
    'enter': [u'\r'],
    'exit': [u'q', u'Q', unichr(27), ],
}


class Selector(AnsiWindow):
    """
    A two-state horizontal lightbar interface.
    """
    # pylint: disable=R0902,R0913,R0904
    #        Too many instance attributes (8/7)
    #        Too many arguments (6/5)
    #        Too many public methods (25/20)

    def __init__(self, yloc, xloc, width, left, right):
        """
        Set screen position of Selector UI and display width of both. The
        highlighted selection is displayed using the self.highlight attribute,
        in order (left, right). The default selection is left.
        """
        self._left = self._selection = left
        self._right = right
        self._moved = False
        self._quit = False
        self._selected = False
        AnsiWindow.__init__(self, 1, width, yloc, xloc)  # height is 1
        self.init_theme()
        self.keyset = VI_KEYSET
        self.init_keystrokes()

    def init_theme(self):
        """
        Initialize colors['selected'] and colors['unselected'].
        """
        from x84.bbs.session import getterminal
        term = getterminal()
        AnsiWindow.init_theme(self)
        self.colors['selected'] = term.reverse
        self.colors['unselected'] = term.normal

    def init_keystrokes(self):
        """
        Merge curses-detected application keys into a VI_KEYSET-formatted
        keyset, for keys 'refresh', 'left', 'right', 'enter', and 'exit'.
        """
        from x84.bbs.session import getterminal
        term = getterminal()
        self.keyset['refresh'].append(term.KEY_REFRESH)
        self.keyset['left'].append(term.KEY_LEFT)
        self.keyset['right'].append(term.KEY_RIGHT)
        self.keyset['enter'].append(term.KEY_ENTER)
        self.keyset['exit'].append(term.KEY_EXIT)

    def process_keystroke(self, keystroke):
        """
        Process a keystroke, toggling self.selection and returning string
        suitable for refresh when changed.
        """
        self._moved = False
        if keystroke in self.keyset['refresh']:
            return self.refresh()
        elif keystroke in self.keyset['left']:
            return self.move_left()
        elif keystroke in self.keyset['right']:
            return self.move_right()
        elif keystroke in self.keyset['toggle']:
            return self.toggle()
        elif keystroke in self.keyset['exit']:
            self._quit = True
        elif keystroke in self.keyset['enter']:
            self._selected = True
        return u''

    def read(self):
        """
        Reads input until the ENTER or ESCAPE key is pressed (Blocking).
        Allows backspacing. Returns unicode text, or None when cancelled.
        """
        from x84.bbs import getch
        from x84.bbs.output import echo
        self._selected = False
        self._quit = False
        echo(self.refresh())
        while not (self.selected or self.quit):
            echo(self.process_keystroke(getch()) or u'')
        if self.quit:
            return None
        return self.selection

    @property
    def selected(self):
        """
        Returns True when keyset['enter'] key detected in process_keystroke
        """
        return self._selected

    @selected.setter
    def selected(self, value):
        # pylint: disable=C0111
        #         Missing docstring
        # this setter should only be used to reset to 'False' for recycling
        assert type(value) is bool
        self._selected = value

    @property
    def selection(self):
        """
        Current selection.
        """
        return self._selection

    @selection.setter
    def selection(self, value):
        # pylint: disable=C0111
        #         Missing docstring
        assert value in (self._left, self._right)
        if self._selection != value:
            self._moved = True
            self._selection = value

    @property
    def left(self):
        """
        Left-side value
        """
        return self._left

    @left.setter
    def left(self, value):
        # pylint: disable=C0111
        #         Missing docstring
        self._left = value

    @property
    def right(self):
        """
        Right-side value
        """
        return self._right

    @right.setter
    def right(self, value):
        # pylint: disable=C0111
        #         Missing docstring
        self._right = value

    def refresh(self):
        """
        Return terminal sequence suitable for re-drawing left/right menubar.
        """
        import math
        import x84.bbs.session
        term = x84.bbs.session.getterminal()
        attr_l = (self.colors.get('selected', u'')
                  if self.selection == self.left
                  else self.colors.get('unselected'))
        attr_r = (self.colors.get('selected', u'')
                  if self.selection == self.right
                  else self.colors.get('unselected'))
        return u''.join((
            self.pos(0, 0), term.normal, attr_l,
            self.left.center(int(math.ceil(float(self.width) / 2))),
            term.normal, attr_r,
            self.right.center(int(math.floor(float(self.width) / 2))),
            term.normal,))

    def move_right(self):
        """
        Force state to right, returning unicode string suitable for refresh.
        If state is unchanged, an empty string is returned.
        """
        if self.selection != self.right:
            self.selection = self.right
            return self.refresh()
        return u''

    def move_left(self):
        """
        Force state to left, returning unicode string suitable for refresh.
        If state is unchanged, an empty string is returned.
        """
        if self.selection != self.left:
            self.selection = self.left
            return self.refresh()
        return u''

    def toggle(self):
        """
        Toggle selection and return unicode string suitable for refresh.
        """
        if self.selection == self.left:
            self.selection = self.right
        else:
            self.selection = self.left
        return self.refresh()

    @property
    def quit(self):
        """
        Returns: True if a terminating or quit character was handled by
        process_keystroke(), such as the escape key, or 'q' by default.
        """
        return self._quit



import shlex
import os

# example dosemu.conf
#  $_cpu = "80486"
#  $_cpu_emu = "vm86"
#  $_external_char_set = "utf8"
#  $_internal_char_set = "cp437"
#  $_term_updfreq = (8)
#  $_layout = "us"
#  $_rawkeyboard = (0)

# example <home>/.dosemu/drive_c/autoexec.bat
#  @echo off
#  path d:\bin;d:\gnu;d:\dosemu
#  set TEMP=c:\tmp
#  prompt $P$G
#  C:\BNU\BNU.COM /L0:57600,8N1 /F
#  lredir.com x: linux\fs\DOS\X
#  unix -e

# notes:
# bnu.com is a fossil driver


# doorfiles configured for DORINFOn.DEF support in LORDCFG.exe

def main():
    from x84.bbs import getsession, getterminal, echo, ini
    from x84.bbs import DOSDoor, Dropfile

    assert ini.CFG.getboolean('dosemu', 'enabled'), (
        'lord.py called but dosemu not enabled in configuration')
    session, term = getsession(), getterminal()

    # dosemu is a bear; as the BBS typically runs as user 'nobody',
    # the $HOME path is often unset or '/'; by exporting an ENV
    # variable 'HOME' to '/DOS', we can manipulate the bbs door
    # dosemu environment with the /DOS/.dosemu folder, and test with
    # 'sudo su - nobody', then launching 'dosemu'
    bin = ini.CFG.get('dosemu', 'bin')
    home = ini.CFG.get('dosemu', 'home')
    lord_path = ini.CFG.get('dosemu', 'lord_path')
    lord_dropfile = ini.CFG.get('dosemu', 'lord_dropfile').upper()
    lord_args = ini.CFG.get('dosemu', 'lord_args')

    assert lord_path != 'no'
    assert os.path.exists(lord_path)
    assert os.path.exists(bin)
    assert os.path.exists(home)

    store_cols, store_rows = None, None
    want_cols, want_rows = 80, 25
    if want_cols != term.width or want_rows != term.height:
        echo(u'\x1b[8;%d;%dt' % (want_rows, want_cols,))
    disp = 1
    while not (term.width == want_cols
            and term.height == want_cols):
        if disp:
            echo(term.bold_blue('\r\n^\r\n'))
            echo(term.bold_blue('\r\n'.join([u'|'] * (want_rows - 3))))
            echo('\r\n')
            echo(term.bold_blue(u'|' + (u'=' * 78) + u'|\r\n'))
            echo(u'for best "screen output", please '
                'resize window to %s x %s (or press return).' % (
                    want_cols, want_rows,))
            disp = 0
        ret = term.inkey(2)
        if ret in (term.KEY_ENTER, u'\r', u'\n'):
            break

    if term.width != want_cols or term.height != want_rows:
        echo('Your dimensions: %s by %s; emulating %s by %s !' % (
            term.width, term.height, want_cols, want_rows,))
        # hand-hack, its ok ... really
        store_cols, store_rows = term.width, term.height
        term.columns, term.rows= want_cols, want_rows
        term.inkey(1)

    session.activity = 'Playing LORD'
    lord_args = lord_args.replace('%#', str(session.node))
    Dropfile(getattr(Dropfile, lord_dropfile)).save(lord_path)
    door = DOSDoor(bin, args=shlex.split(lord_args),
            env_home=home)
    door.run()
    if not (store_cols is None and store_rows is None):
        echo('Restoring dimensions to %s by %s !' % (store_cols, store_rows))
        term.rows, term.columns = store_rows, store_cols
